/**
 *  '$RCSfile: DocumentImpl.java,v $'
 *    Purpose: A Class that represents an XML document
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: 1.5.0
 *
 *   '$Author: sgarg $'
 *     '$Date: 2005/04/15 02:44:07 $'
 * '$Revision: 1.153 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.net.URL;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Vector;

import edu.ucsb.nceas.dbadapter.AbstractDatabase;

import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * A class that represents an XML document. It can be created with a simple
 * document identifier from a database connection. It also will write an XML
 * text document to a database connection using SAX.
 */
public class DocumentImpl
{
    /* Constants */
    public static final String SCHEMA = "Schema";
    public static final String DTD = "DTD";
    public static final String EML200 = "eml200";
    public static final String EML210 = "eml210";
    public static final String EXTERNALSCHEMALOCATIONPROPERTY = "http://apache.org/xml/properties/schema/external-schemaLocation";

    /*
     * public static final String EXTERNALSCHEMALOCATION =
     * "eml://ecoinformatics.org/eml-2.0.0
     * http://dev.nceas.ucsb.edu/tao/schema/eml.xsd"+ "
     * http://www.xml-cml.org/schema/stmml
     * http://dev.nceas.ucsb.edu/tao/schema/stmml.xsd";
     */
    public static final String DECLARATIONHANDLERPROPERTY = "http://xml.org/sax/properties/declaration-handler";
    public static final String LEXICALPROPERTY = "http://xml.org/sax/properties/lexical-handler";
    public static final String VALIDATIONFEATURE = "http://xml.org/sax/features/validation";
    public static final String SCHEMAVALIDATIONFEATURE = "http://apache.org/xml/features/validation/schema";
    public static final String NAMESPACEFEATURE = "http://xml.org/sax/features/namespaces";
    public static final String NAMESPACEPREFIXESFEATURE = "http://xml.org/sax/features/namespace-prefixes";
    public static final String EML2_1_0NAMESPACE = MetaCatUtil
            .getOption("eml2_1_0namespace");
    // "eml://ecoinformatics.org/eml-2.1.0";
    public static final String EML2_0_1NAMESPACE = MetaCatUtil
            .getOption("eml2_0_1namespace");
    // "eml://ecoinformatics.org/eml-2.0.1";
    public static final String EML2_0_0NAMESPACE = MetaCatUtil
            .getOption("eml2_0_0namespace");
    // "eml://ecoinformatics.org/eml-2.0.0";
    public static final String DOCNAME = "docname";
    public static final String PUBLICID = "publicid";
    public static final String SYSTEMID = "systemid";
    static final int ALL = 1;
    static final int WRITE = 2;
    static final int READ = 4;
    private static final AbstractDatabase dbAdapter = MetaCatUtil.dbAdapter;
    private DBConnection connection = null;
    private String docid = null;
    private String updatedVersion = null;
    private String docname = null;
    private String doctype = null;
    private String validateType = null; //base on dtd or schema
    private String createdate = null;
    private String updatedate = null;
    private String system_id = null;
    private String userowner = null;
    private String userupdated = null;
    private int rev;
    private int serverlocation;
    private String docHomeServer;
    private String publicaccess;
    private long rootnodeid;
    private ElementNode rootNode = null;
    private TreeSet nodeRecordList = null;

    /**
     * Constructor used to create a document and read the document information
     * from the database. If readNodes is false, then the node data is not read
     * at this time, but is deferred until it is needed (such as when a call to
     * toXml() is made).
     *
     * @param conn
     *            the database connection from which to read the document
     * @param docid
     *            the identifier of the document to be created
     * @param readNodes
     *            flag indicating whether the xmlnodes should be read
     */
    public DocumentImpl(String docid, boolean readNodes) throws McdbException
    {
        try {
            //this.conn = conn;
            this.docid = docid;

            // Look up the document information
            getDocumentInfo(docid);

            if (readNodes) {
                // Download all of the document nodes using a single SQL query
                // The sort order of the records is determined by the
                // NodeComparator
                // class, and needs to represent a depth-first traversal for the
                // toXml() method to work properly
                nodeRecordList = getNodeRecordList(rootnodeid);
            }

        } catch (McdbException ex) {
            throw ex;
        } catch (Throwable t) {
            throw new McdbException("Error reading document: " + docid);
        }
    }

    /**
     * Constructor, creates document from database connection, used for reading
     * the document
     *
     * @param conn
     *            the database connection from which to read the document
     * @param docid
     *            the identifier of the document to be created
     */
    public DocumentImpl(String docid) throws McdbException
    {
        this(docid, true);
    }

    /**
     * Construct a new document instance, writing the contents to the database.
     * This method is called from DBSAXHandler because we need to know the root
     * element name for documents without a DOCTYPE before creating it.
     *
     * In this constructor, the docid is without rev. There is a string rev to
     * specify the revision user want to upadate. The revion is only need to be
     * greater than current one. It is not need to be sequent number just after
     * current one. So it is only used in update action
     *
     * @param conn
     *            the JDBC Connection to which all information is written
     * @param rootnodeid -
     *            sequence id of the root node in the document
     * @param docname -
     *            the name of DTD, i.e. the name immediately following the
     *            DOCTYPE keyword ( should be the root element name ) or the
     *            root element name if no DOCTYPE declaration provided (Oracle's
     *            and IBM parsers are not aware if it is not the root element
     *            name)
     * @param doctype -
     *            Public ID of the DTD, i.e. the name immediately following the
     *            PUBLIC keyword in DOCTYPE declaration or the docname if no
     *            Public ID provided or null if no DOCTYPE declaration provided
     * @param docid
     *            the docid to use for the UPDATE, no version number
     * @param version,
     *            need to be update
     * @param action
     *            the action to be performed (INSERT OR UPDATE)
     * @param user
     *            the user that owns the document
     * @param pub
     *            flag for public "read" access on document
     * @param serverCode
     *            the serverid from xml_replication on which this document
     *            resides.
     *
     */
    public DocumentImpl(DBConnection conn, long rootNodeId, String docName,
            String docType, String docId, String newRevision, String action,
            String user, String pub, String catalogId, int serverCode)
            throws SQLException, Exception
    {
        this.connection = conn;
        this.rootnodeid = rootNodeId;
        this.docname = docName;
        this.doctype = docType;
        this.docid = docId;
        this.updatedVersion = newRevision;
        writeDocumentToDB(action, user, pub, catalogId, serverCode);
    }

    /**
     * This method will be call in handleUploadRequest in MetacatServlet class
     */
    public static void registerDocument(String docname, String doctype,
            String accnum, String user, String[] groupnames) throws SQLException,
            AccessionNumberException, Exception
    {
        try {
            // get server location for this doc
            int serverLocation = getServerLocationNumber(accnum);
            registerDocument(docname, doctype, accnum, user, groupnames,
                             serverLocation);
        } catch (Exception e) {
            throw e;
        }
    }

    /**
     * Register a document that resides on the filesystem with the database.
     * (ie, just an entry in xml_documents, nothing in xml_nodes). Creates a
     * reference to a filesystem document (used for non-xml data files). This
     * class only be called in MetaCatServerlet.
     *
     * @param conn
     *            the JDBC Connection to which all information is written
     * @param docname -
     *            the name of DTD, i.e. the name immediately following the
     *            DOCTYPE keyword ( should be the root element name ) or the
     *            root element name if no DOCTYPE declaration provided (Oracle's
     *            and IBM parsers are not aware if it is not the root element
     *            name)
     * @param doctype -
     *            Public ID of the DTD, i.e. the name immediately following the
     *            PUBLIC keyword in DOCTYPE declaration or the docname if no
     *            Public ID provided or null if no DOCTYPE declaration provided
     * @param accnum
     *            the accession number to use for the INSERT OR UPDATE, which
     *            includes a revision number for this revision of the document
     *            (e.g., knb.1.1)
     * @param user
     *            the user that owns the document
     * @param groupnames
     *            the groups that owns the document
     * @param serverCode
     *            the serverid from xml_replication on which this document
     *            resides.
     */
    public static void registerDocument(String docname, String doctype,
            String accnum, String user, String[] groups, int serverCode)
            throws SQLException, AccessionNumberException, Exception
    {

        DBConnection dbconn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;
        //MetaCatUtil util = new MetaCatUtil();
        AccessionNumber ac;
        String action = null;
        try {
            //dbconn = util.openDBConnection();
            //check out DBConnection
            dbconn = DBConnectionPool
                    .getDBConnection("DocumentImpl.registerDocument");
            serialNumber = dbconn.getCheckOutSerialNumber();
            String docIdWithoutRev = MetaCatUtil.getDocIdFromString(accnum);
            int userSpecifyRev = MetaCatUtil.getVersionFromString(accnum);
            int revInDataBase = getLatestRevisionNumber(docIdWithoutRev);
            //revIndataBase=-1, there is no record in xml_documents table
            //the data file is a new one, inert it into table
            //user specified rev should be great than 0
            if (revInDataBase == -1 && userSpecifyRev > 0) {
                ac = new AccessionNumber(accnum, "insert");
                action = "insert";
            }
            //rev is greater the last revsion number and revInDataBase isn't -1
            // it is a updated data file
            else if (userSpecifyRev > revInDataBase && revInDataBase > 0) {

                if (!hasWritePermission(user, groups, accnum)) { throw new Exception(
                   "User " + user
                   + " does not have permission to update the document"
                   + accnum); }

                //archive the old entry
                archiveDocRevision(docIdWithoutRev, user);
                //delete the old entry in xml_documents
                //deleteXMLDocuments(docIdWithoutRev);
                ac = new AccessionNumber(accnum, "update");
                action = "update";
            }
            //other situation
            else {

                throw new Exception("Revision number couldn't be "
                        + userSpecifyRev);
            }
            String docid = ac.getDocid();
            String rev = ac.getRev();
            /*
             * SimpleDateFormat formatter = new SimpleDateFormat ("MM/dd/yy
             * HH:mm:ss"); Date localtime = new Date(); String dateString =
             * formatter.format(localtime); String
             * sqlDateString=dbAdapter.toDate(dateString, "MM/DD/YY
             * HH24:MI:SS");
             */
            String sqlDateString = dbAdapter.getDateTimeFunction();

            StringBuffer sql = new StringBuffer();
            if (action != null && action.equals("insert")) {
                sql.append("insert into xml_documents (docid, docname, " +
                           "doctype, ");
                sql.append("user_owner, user_updated, server_location, " +
                           "rev,date_created");
                sql.append(", date_updated, public_access) values ('");
                sql.append(docid).append("','");
                sql.append(docname).append("','");
                sql.append(doctype).append("','");
                sql.append(user).append("','");
                sql.append(user).append("','");
                sql.append(serverCode).append("','");
                sql.append(rev).append("',");
                sql.append(sqlDateString).append(",");
                sql.append(sqlDateString).append(",");
                sql.append("'0')");
            } else if (action != null && action.equals("update")) {
                sql.append("update xml_documents set docname ='");
                sql.append(docname).append("', ");
                sql.append("user_updated='");
                sql.append(user).append("', ");
                sql.append("server_location='");
                sql.append(serverCode).append("',");
                sql.append("rev='");
                sql.append(rev).append("',");
                sql.append("date_updated=");
                sql.append(sqlDateString);
                sql.append(" where docid='");
                sql.append(docid).append("'");
            }
            pstmt = dbconn.prepareStatement(sql.toString());
            pstmt.execute();
            pstmt.close();
            //dbconn.close();
        } finally {
            try {
                if(pstmt != null){
                    pstmt.close();
                }
            } finally {
                DBConnectionPool.returnDBConnection(dbconn, serialNumber);
            }
        }
    }

    /**
     * Register a document that resides on the filesystem with the database.
     * (ie, just an entry in xml_documents, nothing in xml_nodes). Creates a
     * reference to a filesystem document (used for non-xml data files) This
     * method will be called for register data file in xml_documents in
     * Replication. This method is revised from registerDocument.
     *
     * @param conn
     *            the JDBC Connection to which all information is written
     * @param docname -
     *            the name of DTD, i.e. the name immediately following the
     *            DOCTYPE keyword ( should be the root element name ) or the
     *            root element name if no DOCTYPE declaration provided (Oracle's
     *            and IBM parsers are not aware if it is not the root element
     *            name)
     * @param doctype -
     *            Public ID of the DTD, i.e. the name immediately following the
     *            PUBLIC keyword in DOCTYPE declaration or the docname if no
     *            Public ID provided or null if no DOCTYPE declaration provided
     * @param accnum
     *            the accession number to use for the INSERT OR UPDATE, which
     *            includes a revision number for this revision of the document
     *            (e.g., knb.1.1)
     * @param user
     *            the user that owns the document
     * @param serverCode
     *            the serverid from xml_replication on which this document
     *            resides.
     */
    public static void registerDocumentInReplication(String docname,
            String doctype, String accnum, String user, int serverCode)
            throws SQLException, AccessionNumberException, Exception
    {
        DBConnection dbconn = null;
        int serialNumber = -1;
        //MetaCatUtil util = new MetaCatUtil();
        AccessionNumber ac;
        PreparedStatement pstmt = null;
        String action = null;
        try {
            //dbconn = util.openDBConnection();
            dbconn = DBConnectionPool.getDBConnection(
                    "DocumentImpl.registerDocumentInReplication");
            serialNumber = dbconn.getCheckOutSerialNumber();
            String docIdWithoutRev = MetaCatUtil
                    .getDocIdFromAccessionNumber(accnum);
            int userSpecifyRev = MetaCatUtil
                    .getRevisionFromAccessionNumber(accnum);
            int revInDataBase = getLatestRevisionNumber(docIdWithoutRev);
            //revIndataBase=-1, there is no record in xml_documents table
            //the data file is a new one, inert it into table
            //user specified rev should be great than 0
            if (revInDataBase == -1 && userSpecifyRev >= 0) {
                ac = new AccessionNumber(accnum, "insert");
                action = "insert";
            }
            //rev is greater the last revsion number and revInDataBase isn't -1
            // it is a updated data file
            else if (userSpecifyRev > revInDataBase && revInDataBase >= 0) {

                //archive the old entry
                archiveDocRevision(docIdWithoutRev, user);
                //delete the old entry in xml_documents
                //deleteXMLDocuments(docIdWithoutRev);
                ac = new AccessionNumber(accnum, "update");
                action = "update";
            }
            // local server has newer version, then notify the remote server
            else if (userSpecifyRev < revInDataBase && revInDataBase > 0) {
                throw new Exception("Local server: "
                        + MetaCatUtil.getOption("server")
                        + " has newer revision of doc: " + docIdWithoutRev
                        + "." + revInDataBase + ". Please notify it.");
            }
            //other situation
            else {

                throw new Exception("Revision number couldn't be "
                        + userSpecifyRev);
            }
            String docid = ac.getDocid();
            String rev = ac.getRev();
            /*
             * SimpleDateFormat formatter = new SimpleDateFormat ("MM/dd/yy
             * HH:mm:ss"); Date localtime = new Date(); String dateString =
             * formatter.format(localtime); String
             * sqlDateString=dbAdapter.toDate(dateString, "MM/DD/YY
             * HH24:MI:SS");
             */
            String sqlDateString = dbAdapter.getDateTimeFunction();

            StringBuffer sql = new StringBuffer();
            if (action != null && action.equals("insert")) {
                sql.append("insert into xml_documents (docid, docname, " +
                           "doctype, ");
                sql.append("user_owner, user_updated, server_location, " +
                           "rev,date_created");
                sql.append(", date_updated, public_access) values ('");
                sql.append(docid).append("','");
                sql.append(docname).append("','");
                sql.append(doctype).append("','");
                sql.append(user).append("','");
                sql.append(user).append("','");
                sql.append(serverCode).append("','");
                sql.append(rev).append("',");
                sql.append(sqlDateString).append(",");
                sql.append(sqlDateString).append(",");
                sql.append("'0')");
            } else if (action != null && action.equals("update")) {
                sql.append("update xml_documents set docname ='");
                sql.append(docname).append("', ");
                sql.append("user_updated='");
                sql.append(user).append("', ");
                sql.append("server_location='");
                sql.append(serverCode).append("',");
                sql.append("rev='");
                sql.append(rev).append("',");
                sql.append("date_updated=");
                sql.append(sqlDateString);
                sql.append(" where docid='");
                sql.append(docid).append("'");
            }
            // Set auto commit fasle
            dbconn.setAutoCommit(false);
            pstmt = dbconn.prepareStatement(sql.toString());

            pstmt.execute();
            // Commit the insert
            dbconn.commit();
            pstmt.close();
            //dbconn.close();
        } finally {
            // Set DBConnection auto commit true
            dbconn.setAutoCommit(true);
            pstmt.close();
            DBConnectionPool.returnDBConnection(dbconn, serialNumber);
        }
    }

    /**
     * This method will register a data file entry in xml_documents and save a
     * data file input Stream into file system.. It is only used in replication
     *
     * @param input,
     *            the input stream which contain the file content.
     * @param ,
     *            the input stream which contain the file content
     * @param docname -
     *            the name of DTD, for data file, it is a docid number.
     * @param doctype -
     *            "BIN" for data file
     * @param accnum
     *            the accession number to use for the INSERT OR UPDATE, which
     *            includes a revision number for this revision of the document
     *            (e.g., knb.1.1)
     * @param user
     *            the user that owns the document
     * @param docHomeServer,
     *            the home server of the docid
     * @param notificationServer,
     *            the server to notify force replication info to local metacat
     */
    public static void writeDataFileInReplication(InputStream input,
            String filePath, String docname, String doctype, String accnum,
            String user, String docHomeServer, String notificationServer)
            throws SQLException, AccessionNumberException, Exception
    {
        int serverCode = -2;

        if (filePath == null || filePath.equals("")) { throw new Exception(
                "Please specify the directory where file will be store"); }
        if (accnum == null || accnum.equals("")) { throw new Exception(
                "Please specify the stored file name"); }

        // If server is not int the xml replication talbe, insert it into
        // xml_replication table
        //serverList.addToServerListIfItIsNot(docHomeServer);
        insertServerIntoReplicationTable(docHomeServer);

        // Get server code again
        serverCode = getServerCode(docHomeServer);

        //register data file into xml_documents table
        registerDocumentInReplication(docname, doctype, accnum, user,
                serverCode);
        //write inputstream into file system.
        File dataDirectory = new File(filePath);
        File newFile = new File(dataDirectory, accnum);

        // create a buffered byte output stream
        // that uses a default-sized output buffer
        FileOutputStream fos = new FileOutputStream(newFile);
        BufferedOutputStream outPut = new BufferedOutputStream(fos);

        BufferedInputStream bis = null;
        bis = new BufferedInputStream(input);
        byte[] buf = new byte[4 * 1024]; // 4K buffer
        int b = bis.read(buf);

        while (b != -1) {
            outPut.write(buf, 0, b);
            b = bis.read(buf);
        }
        bis.close();
        outPut.close();
        fos.close();

        // Force replicate data file
        ForceReplicationHandler forceReplication = new ForceReplicationHandler(
                accnum, false, notificationServer);
    }

    /**
     * Get a lock for a given document.
     */
    public static boolean getDataFileLockGrant(String accnum) throws Exception
    {
        try {
            int serverLocation = getServerLocationNumber(accnum);
            return getDataFileLockGrant(accnum, serverLocation);
        } catch (Exception e) {
            throw e;
        }
    }

    /**
     * The method will check if metacat can get data file lock grant If server
     * code is 1, it get. If server code is not 1 but call replication getlock
     * successfully, it get else, it didn't get
     *
     * @param accnum,
     *            the ID of the document
     * @param action,
     *            the action to the document
     * @param serverCode,
     *            the server location code
     */
    public static boolean getDataFileLockGrant(String accnum, int serverCode)
            throws Exception
    {
        boolean flag = true;
        //MetaCatUtil util = new MetaCatUtil();
        String docid = MetaCatUtil.getDocIdFromString(accnum);
        int rev = MetaCatUtil.getVersionFromString(accnum);

        if (serverCode == 1) {
            flag = true;
            return flag;
        }

        //if((serverCode != 1 && action.equals("UPDATE")) )
        if (serverCode != 1) { //if this document being written is not a
                               // resident of this server then
            //we need to try to get a lock from it's resident server. If the
            //resident server will not give a lock then we send the user a
            // message
            //saying that he/she needs to download a new copy of the file and
            //merge the differences manually.

            String server = MetacatReplication
                    .getServerNameForServerCode(serverCode);
            MetacatReplication.replLog("attempting to lock " + accnum);
            URL u = new URL("https://" + server + "?server="
                    + MetaCatUtil.getLocalReplicationServerName()
                    + "&action=getlock&updaterev=" + rev + "&docid=" + docid);
            //System.out.println("sending message: " + u.toString());
            String serverResStr = MetacatReplication.getURLContent(u);
            String openingtag = serverResStr.substring(0, serverResStr
                    .indexOf(">") + 1);
            if (openingtag.equals("<lockgranted>")) {
                //the lock was granted go ahead with the insert
                //System.out.println("In lockgranted");
                MetacatReplication.replLog("lock granted for " + accnum
                        + " from " + server);
                flag = true;
                return flag;
            }//if

            else if (openingtag.equals("<filelocked>")) {//the file is
                                                         // currently locked by
                                                         // another user
                //notify our user to wait a few minutes, check out a new copy
                // and try
                //again.
                //System.out.println("file locked");
                MetacatReplication.replLog("lock denied for " + accnum + " on "
                        + server + " reason: file already locked");
                throw new Exception(
                        "The file specified is already locked by another "
                                + "user.  Please wait 30 seconds, checkout the "
                                + "newer document, merge your changes and try "
                                + "again.");
            } else if (openingtag.equals("<outdatedfile>")) {//our file is
                                                             // outdated. notify
                                                             // our user to
                                                             // check out a new
                                                             // copy of the
                //file and merge his version with the new version.
                //System.out.println("outdated file");
                MetacatReplication.replLog("lock denied for " + accnum + " on "
                        + server + " reason: local file outdated");
                throw new Exception(
                        "The file you are trying to update is an outdated"
                         + " version.  Please checkout the newest document, "
                         + "merge your changes and try again.");
            }
        }
        return flag;
    }

    /**
     * get the document name
     */
    public String getDocname()
    {
        return docname;
    }

    /**
     * get the document type (which is the PublicID)
     */
    public String getDoctype()
    {
        return doctype;
    }

    /**
     * get the system identifier
     */
    public String getSystemID()
    {
        return system_id;
    }

    /**
     * get the root node identifier
     */
    public long getRootNodeID()
    {
        return rootnodeid;
    }

    /**
     * get the creation date
     */
    public String getCreateDate()
    {
        return createdate;
    }

    /**
     * get the update date
     */
    public String getUpdateDate()
    {
        return updatedate;
    }

    /**
     * Get the document identifier (docid)
     */
    public String getDocID()
    {
        return docid;
    }

    public String getUserowner()
    {
        return userowner;
    }

    public String getUserupdated()
    {
        return userupdated;
    }

    public int getServerlocation()
    {
        return serverlocation;
    }

    public String getDocHomeServer()
    {
        return docHomeServer;
    }

    public String getPublicaccess()
    {
        return publicaccess;
    }

    public int getRev()
    {
        return rev;
    }

    public String getValidateType()
    {
        return validateType;
    }

    /**
     * Print a string representation of the XML document
     */
    public String toString(String user, String[] groups, boolean withInlinedata)
    {
        StringWriter docwriter = new StringWriter();
        try {
            this.toXml(docwriter, user, groups, withInlinedata);
        } catch (McdbException mcdbe) {
            return null;
        }
        String document = docwriter.toString();
        return document;
    }

    /**
     * Print a string representation of the XML document
     */
    public String toString()
    {
        StringWriter docwriter = new StringWriter();
        String userName = null;
        String[] groupNames = null;
        boolean withInlineData = true;
        try {
            this.toXml(docwriter, userName, groupNames, withInlineData);
        } catch (McdbException mcdbe) {
            return null;
        }
        String document = docwriter.toString();
        return document;
    }

    /**
     * Get a text representation of the XML document as a string This older
     * algorithm uses a recursive tree of Objects to represent the nodes of the
     * tree. Each object is passed the data for the document and searches all of
     * the document data to find its children nodes and recursively build. Thus,
     * because each node reads the whole document, this algorithm is extremely
     * slow for larger documents, and the time to completion is O(N^N) wrt the
     * number of nodes. See toXml() for a better algorithm.
     */
    public String readUsingSlowAlgorithm() throws McdbException
    {
        StringBuffer doc = new StringBuffer();

        // First, check that we have the needed node data, and get it if not
        if (nodeRecordList == null) {
            nodeRecordList = getNodeRecordList(rootnodeid);
        }

        // Create the elements from the downloaded data in the TreeSet
        rootNode = new ElementNode(nodeRecordList, rootnodeid);

        // Append the resulting document to the StringBuffer and return it
        doc.append("<?xml version=\"1.0\"?>\n");

        if (docname != null) {
            if ((doctype != null) && (system_id != null)) {
                doc.append("<!DOCTYPE " + docname + " PUBLIC \"" + doctype
                        + "\" \"" + system_id + "\">\n");
            } else {
                doc.append("<!DOCTYPE " + docname + ">\n");
            }
        }
        doc.append(rootNode.toString());

        return (doc.toString());
    }

    /**
     * Print a text representation of the XML document to a Writer
     *
     * @param pw
     *            the Writer to which we print the document Now we decide no
     *            matter withinInlineData's value, the document will
     *
     */
    public void toXml(Writer pw, String user, String[] groups,
            boolean withInLineData) throws McdbException
    {
        // flag for process eml2
        boolean proccessEml2 = false;
        boolean storedDTD = false;//flag to inidate publicid or system
        // id stored in db or not
        boolean firstElement = true;
        String dbDocName = null;
        String dbPublicID = null;
        String dbSystemID = null;

        if (doctype != null
                && (doctype.equals(EML2_0_0NAMESPACE)
                        || doctype.equals(EML2_0_1NAMESPACE) || doctype
                        .equals(EML2_1_0NAMESPACE))) {
            proccessEml2 = true;
        }
        // flag for process inline data
        boolean prcocessInlineData = false;

        TreeSet nodeRecordLists = null;
        PrintWriter out = null;
        if (pw instanceof PrintWriter) {
            out = (PrintWriter) pw;
        } else {
            out = new PrintWriter(pw);
        }

        MetaCatUtil util = new MetaCatUtil();

        // Here add code to handle subtree access control
        /*
         * PermissionController control = new PermissionController(docid);
         * Hashtable unaccessableSubTree =control.hasUnaccessableSubTree(user,
         * groups, AccessControlInterface.READSTRING);
         *
         * if (!unaccessableSubTree.isEmpty()) {
         *
         * nodeRecordLists = getPartNodeRecordList(rootnodeid,
         * unaccessableSubTree);
         *  } else { nodeRecordLists = getNodeRecordList(rootnodeid); }
         */
        nodeRecordLists = getNodeRecordList(rootnodeid);
        Stack openElements = new Stack();
        boolean atRootElement = true;
        boolean previousNodeWasElement = false;

        // Step through all of the node records we were given

        Iterator it = nodeRecordLists.iterator();

        while (it.hasNext()) {

            NodeRecord currentNode = (NodeRecord) it.next();
            util.debugMessage("[Got Node ID: " + currentNode.nodeid + " ("
                    + currentNode.parentnodeid + ", " + currentNode.nodeindex
                    + ", " + currentNode.nodetype + ", " + currentNode.nodename
                    + ", " + currentNode.nodedata + ")]", 50);
            // Print the end tag for the previous node if needed
            //
            // This is determined by inspecting the parent nodeid for the
            // currentNode. If it is the same as the nodeid of the last element
            // that was pushed onto the stack, then we are still in that
            // previous
            // parent element, and we do nothing. However, if it differs, then
            // we
            // have returned to a level above the previous parent, so we go into
            // a loop and pop off nodes and print out their end tags until we
            // get
            // the node on the stack to match the currentNode parentnodeid
            //
            // So, this of course means that we rely on the list of elements
            // having been sorted in a depth first traversal of the nodes, which
            // is handled by the NodeComparator class used by the TreeSet
            if (!atRootElement) {
                NodeRecord currentElement = (NodeRecord) openElements.peek();
                if (currentNode.parentnodeid != currentElement.nodeid) {
                    while (currentNode.parentnodeid != currentElement.nodeid) {
                        currentElement = (NodeRecord) openElements.pop();
                        util.debugMessage("\n POPPED: "
                                + currentElement.nodename, 50);
                        if (previousNodeWasElement) {
                            out.print(">");
                            previousNodeWasElement = false;
                        }
                        if (currentElement.nodeprefix != null) {
                            out.print("</" + currentElement.nodeprefix + ":"
                                    + currentElement.nodename + ">");
                        } else {
                            out.print("</" + currentElement.nodename + ">");
                        }
                        currentElement = (NodeRecord) openElements.peek();
                    }
                }
            }

            // Handle the DOCUMENT node
            if (currentNode.nodetype.equals("DOCUMENT")) {
                out.print("<?xml version=\"1.0\"?>");

                // Handle the ELEMENT nodes
            } else if (currentNode.nodetype.equals("ELEMENT")) {
                if (atRootElement) {
                    atRootElement = false;
                } else {
                    if (previousNodeWasElement) {
                        out.print(">");
                    }
                }

                // if publicid or system is not stored into db send it out by
                // default
                if (!storedDTD & firstElement) {
                    if (docname != null && validateType != null
                            && validateType.equals(DTD)) {
                        if ((doctype != null) && (system_id != null)) {

                            out.print("<!DOCTYPE " + docname + " PUBLIC \""
                                    + doctype + "\" \"" + system_id + "\">");
                        } else {

                            out.print("<!DOCTYPE " + docname + ">");
                        }
                    }
                }
                firstElement = false;
                openElements.push(currentNode);
                util.debugMessage("\n PUSHED: " + currentNode.nodename, 50);
                previousNodeWasElement = true;
                if (currentNode.nodeprefix != null) {
                    out.print("<" + currentNode.nodeprefix + ":"
                            + currentNode.nodename);
                } else {
                    out.print("<" + currentNode.nodename);
                }

                // if currentNode is inline and handle eml2, set flag proccess
                // in
                if (currentNode.nodename != null
                        && currentNode.nodename.equals(Eml200SAXHandler.INLINE)
                        && proccessEml2) {
                    prcocessInlineData = true;
                }

                // Handle the ATTRIBUTE nodes
            } else if (currentNode.nodetype.equals("ATTRIBUTE")) {
                if (currentNode.nodeprefix != null) {
                    out.print(" " + currentNode.nodeprefix + ":"
                            + currentNode.nodename + "=\""
                            + currentNode.nodedata + "\"");
                } else {
                    out.print(" " + currentNode.nodename + "=\""
                            + currentNode.nodedata + "\"");
                }

                // Handle the NAMESPACE nodes
            } else if (currentNode.nodetype.equals("NAMESPACE")) {
                out.print(" xmlns:" + currentNode.nodename + "=\""
                        + currentNode.nodedata + "\"");

                // Handle the TEXT nodes
            } else if (currentNode.nodetype.equals("TEXT")) {
                if (previousNodeWasElement) {
                    out.print(">");
                }
                if (!prcocessInlineData) {
                    // if it is not inline data just out put data
                    out.print(currentNode.nodedata);
                } else {
                    // if it is inline data first to get the inline data
                    // internal id
                    String fileName = currentNode.nodedata;
                    String accessfileName = MetaCatUtil
                            .getDocIdWithoutRevFromInlineDataID(fileName);
                    // check if user has read permision for this inline data
                    boolean readInlinedata = false;
                    try {
                        Hashtable unReadableInlineDataList =
                            PermissionController
                                .getUnReadableInlineDataIdList(accessfileName,
                                        user, groups, true);
                        if (!unReadableInlineDataList.containsValue(fileName)) {
                            readInlinedata = true;
                        }
                    } catch (Exception e) {
                        throw new McdbException(e.getMessage());
                    }

                    if (readInlinedata) {
                        //user want to see it, pull out from
                        // file system and output it
                        // for inline data, the data base only store the file
                        // name, so we
                        // can combine the file name and inline data file path,
                        // to get it

                        Reader reader = Eml200SAXHandler
                                .readInlineDataFromFileSystem(fileName);
                        char[] characterArray = new char[4 * 1024];
                        try {
                            int length = reader.read(characterArray);
                            while (length != -1) {
                                out
                                        .print(new String(characterArray, 0,
                                                length));
                                out.flush();
                                length = reader.read(characterArray);
                            }
                            reader.close();
                        } catch (IOException e) {
                            throw new McdbException(e.getMessage());
                        }
                    }//if can read inline data
                    else {
                        // if user can't read it, we only send it back a empty
                        // string
                        // in inline element.
                        out.print("");
                    }// else can't read inlinedata
                    // reset proccess inline data false
                    prcocessInlineData = false;
                }// in inlinedata part
                previousNodeWasElement = false;
                // Handle the COMMENT nodes
            } else if (currentNode.nodetype.equals("COMMENT")) {
                if (previousNodeWasElement) {
                    out.print(">");
                }
                out.print("<!--" + currentNode.nodedata + "-->");
                previousNodeWasElement = false;

                // Handle the PI nodes
            } else if (currentNode.nodetype.equals("PI")) {
                if (previousNodeWasElement) {
                    out.print(">");
                }
                out.print("<?" + currentNode.nodename + " "
                        + currentNode.nodedata + "?>");
                previousNodeWasElement = false;
                // Handle the DTD nodes (docname, publicid, systemid)
            } else if (currentNode.nodetype.equals(DTD)) {
                storedDTD = true;
                if (currentNode.getNodeName().equals(DOCNAME)) {
                    dbDocName = currentNode.getNodeData();
                }
                if (currentNode.getNodeName().equals(PUBLICID)) {
                    dbPublicID = currentNode.getNodeData();
                }
                if (currentNode.getNodeName().equals(SYSTEMID)) {
                    dbSystemID = currentNode.getNodeData();
                    // send out <!doctype .../>
                    if (dbDocName != null) {
                        if ((dbPublicID != null) && (dbSystemID != null)) {

                            out
                                    .print("<!DOCTYPE " + dbDocName
                                            + " PUBLIC \"" + dbPublicID
                                            + "\" \"" + dbSystemID + "\">");
                        } else {

                            out.print("<!DOCTYPE " + dbDocName + ">");
                        }
                    }

                    //reset these variable
                    dbDocName = null;
                    dbPublicID = null;
                    dbSystemID = null;
                }

                // Handle any other node type (do nothing)
            } else {
                // Any other types of nodes are not handled.
                // Probably should throw an exception here to indicate this
            }
            out.flush();
        }

        // Print the final end tag for the root element
        while (!openElements.empty()) {
            NodeRecord currentElement = (NodeRecord) openElements.pop();
            util.debugMessage("\n POPPED: " + currentElement.nodename, 50);
            if (currentElement.nodeprefix != null) {
                out.print("</" + currentElement.nodeprefix + ":"
                        + currentElement.nodename + ">");
            } else {
                out.print("</" + currentElement.nodename + ">");
            }
        }
        out.flush();
    }

    /**
     * Build the index records for this document.  For each node, all absolute
     * and relative paths to the root of the document are created and inserted
     * into the xml_index table.  This requires that the DocumentImpl instance
     * exists, so first call the constructor that reads the document from the
     * database.
     *
     * @throws McdbException on error getting the node records for the document
     */
    public void buildIndex() throws McdbException
    {
        MetaCatUtil util = new MetaCatUtil();
        TreeSet nodeRecordLists = getNodeRecordList(rootnodeid);
        Stack openElements = new Stack();
        boolean atRootElement = true;
        long rootNodeId = -1;

        // Build a map of the same records that are present in the
        // TreeSet so that any node can be easily accessed by nodeId
        HashMap nodeRecordMap = new HashMap();
        Iterator it = nodeRecordLists.iterator();
        while (it.hasNext()) {
            NodeRecord currentNode = (NodeRecord) it.next();
            Long nodeId = new Long(currentNode.getNodeId());
            nodeRecordMap.put(nodeId, currentNode);
        }

        // Opening separate db connection for deleting and writing
        // XML Index -- be sure that it is all in one db transaction
        int serialNumber = -1;
        DBConnection dbConn = null;
        try {
            dbConn = DBConnectionPool.getDBConnection(
                "DocumentImpl.buildIndex");
            serialNumber = dbConn.getCheckOutSerialNumber();
            dbConn.setAutoCommit(false);
            //make sure record is done
            //checkDocumentTable();

            // Delete the previous index entries for this document
            deleteNodeIndex(dbConn);

	    // Delete all the entries in xml_queryresult
            PreparedStatement pstmt = dbConn.prepareStatement(
                    "DELETE FROM xml_queryresult WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            dbConn.increaseUsageCount(1);

	    
            // Step through all of the node records we were given
            // and build the new index and update the database
            it = nodeRecordLists.iterator();
            while (it.hasNext()) {
                NodeRecord currentNode = (NodeRecord) it.next();
                HashMap pathList = new HashMap();
                if (currentNode.nodetype.equals("ELEMENT") ||
                    currentNode.nodetype.equals("ATTRIBUTE") ) {

                    if (atRootElement) {
                        rootNodeId = currentNode.getNodeId();
                        atRootElement = false;
                    }
                    traverseParents(nodeRecordMap, rootNodeId,
                            currentNode.getNodeId(),
                            currentNode.getNodeId(), "", pathList);
                    updateNodeIndex(dbConn, pathList);
                }
            }
            dbConn.commit();
        } catch (SQLException e) {
            MetaCatUtil.debugMessage(
                "SQL Exception while inserting path index in " +
                "DocumentImpl.buildIndex for document " + docid, 10);
      		MetaCatUtil.debugMessage(e.getMessage(), 10);
            e.printStackTrace();
            try {
                dbConn.rollback();
            } catch (SQLException sqle) {
            	MetaCatUtil.debugMessage(
					"Error while rolling back commit in DocumentImpl.buildIndex"
                    + "\n" + sqle.getMessage(), 10);
            }
        } finally {
			DBConnectionPool.returnDBConnection(dbConn, serialNumber);
        }
    }

    /**
     * Recurse up the parent node hierarchy and add each node to the
     * hashmap of paths to be indexed.
     *
     * @param records the set of records hashed by nodeId
     * @param rootNodeId the id of the root element of the document
     * @param leafNodeId the id of the leafNode being processed
     * @param id the id of the current node to be processed
     * @param children the string representation of all child nodes of this id
     * @param pathList the hash to which paths are added
     * @param nodedata the nodedata for the current node
     */
    private void traverseParents(HashMap records, long rootNodeId,
            long leafNodeId, long id,
            String children, HashMap pathList) {
        Long nodeId = new Long(id);
        NodeRecord current = (NodeRecord)records.get(nodeId);
        long parentId = current.getParentNodeId();
        String currentName = current.getNodeName();
        if (current.nodetype.equals("ELEMENT") ||
            current.nodetype.equals("ATTRIBUTE") ) {

            if (children.equals("")) {
                if (current.nodetype.equals("ATTRIBUTE")) {
                    currentName = "@" + currentName;
                }
                MetaCatUtil.debugMessage("A: " + currentName +"\n", 60);
                pathList.put(currentName, new PathIndexEntry(leafNodeId,
                    currentName, docid, doctype, parentId));
            }
            currentName = "/" + currentName;
            currentName = currentName + children;
            if (parentId != 0) {
                traverseParents(records, rootNodeId, leafNodeId,
                    parentId, currentName, pathList);
            }
            String path = current.getNodeName() + children;
            if (!children.equals("")) {
                MetaCatUtil.debugMessage("B: " + path +"\n", 60);
                pathList.put(path, new PathIndexEntry(leafNodeId, path, docid,
                    doctype, parentId));
            }
            if (id == rootNodeId) {
                String fullPath = '/' + path;
                MetaCatUtil.debugMessage("C: " + fullPath +"\n", 60);
                pathList.put(fullPath, new PathIndexEntry(leafNodeId, fullPath,
                    docid, doctype, parentId));
            }
        }
    }

    /**
     * Delete the paths from the xml_index table on the database in preparation
     * of a subsequent update.
     *
     * @param conn the database connection to use, keeping a single transaction
     * @throws SQLException if there is an error deleting from the db
     */
    private void deleteNodeIndex(DBConnection conn) throws SQLException
    {
        String familyId = MetaCatUtil.getDocIdFromString(docid);
        String sql = "DELETE FROM xml_index WHERE docid LIKE ?";
        MetaCatUtil.debugMessage(sql, 55);
        MetaCatUtil.debugMessage("SQL is: " + sql, 60);

        PreparedStatement pstmt = conn.prepareStatement(sql);

        // Increase usage count for the connection
        conn.increaseUsageCount(1);

        // Execute the delete and close the statement
        pstmt.setString(1, familyId);
        int rows = pstmt.executeUpdate();
        pstmt.close();
        MetaCatUtil.debugMessage("Deleted " + rows + " rows from xml_index " +
            "for document " + docid, 55);
    }

    /**
	 * Insert the paths from the pathList into the xml_index table on the
     * database.
     *
     * @param conn the database connection to use, keeping a single transaction
     * @param pathList the hash of paths to insert
     * @throws SQLException if there is an error inserting into the db
     */
    private void updateNodeIndex(DBConnection conn, HashMap pathList)
    	throws SQLException
    {
        // Create an insert statement to reuse for all of the path
        // insertions
        PreparedStatement pstmt = conn.prepareStatement(
                "INSERT INTO xml_index (nodeid, path, docid, doctype, " +
                "parentnodeid) " + "VALUES (?, ?, ?, ?, ?)");
        // Increase usage count for the connection
        conn.increaseUsageCount(1);
        String familyId = MetaCatUtil.getDocIdFromString(docid);
        pstmt.setString(3, familyId);
        pstmt.setString(4, doctype);

        // Step through the hashtable and insert each of the path values
        Iterator it = pathList.values().iterator();
        while (it.hasNext()) {
            PathIndexEntry entry = (PathIndexEntry)it.next();
            MetaCatUtil.debugMessage("Inserting: " + entry.nodeId +
                " (" + entry.parentId + "): " + entry.path, 60);
            pstmt.setLong(1, entry.nodeId);
            pstmt.setString(2, entry.path);
            pstmt.setLong(5, entry.parentId);
            pstmt.executeUpdate();
        }
        // Close the database statement
        pstmt.close();
    }

    private boolean isRevisionOnly(DocumentIdentifier docid) throws Exception
    {
        //System.out.println("inRevisionOnly");
        DBConnection dbconn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;
        String rev = docid.getRev();
        String newid = docid.getIdentifier();
        try {
            dbconn = DBConnectionPool
                    .getDBConnection("DocumentImpl.isRevisionOnly");
            serialNumber = dbconn.getCheckOutSerialNumber();
            pstmt = dbconn.prepareStatement("select rev from xml_documents "
                    + "where docid like '" + newid + "'");
            pstmt.execute();
            ResultSet rs = pstmt.getResultSet();
            boolean tablehasrows = rs.next();
            if (rev.equals("newest") || rev.equals("all")) { return false; }

            if (tablehasrows) {
                int r = rs.getInt(1);
                pstmt.close();
                if (new Integer(rev).intValue() == r) { //the current revision
                                                        // in in xml_documents
                    //System.out.println("returning false");
                    return false;
                } else if (new Integer(rev).intValue() < r) { //the current
                                                              // revision is in
                                                              // xml_revisions.
                    //System.out.println("returning true");
                    return true;
                } else if (new Integer(rev).intValue() > r) { //error, rev
                                                              // cannot be
                                                              // greater than r
                throw new Exception(
                        "requested revision cannot be greater than "
                                + "the latest revision number."); }
            }
            // Get miss docid and rev, throw to McdDocNotFoundException
            String missDocId = MetaCatUtil.getDocIdFromString(docid.toString());
            String missRevision = MetaCatUtil.getRevisionStringFromString(docid
                    .toString());
            throw new McdbDocNotFoundException("the requested docid '"
                    + docid.toString() + "' does not exist", missDocId,
                    missRevision);
        }//try
        finally {
            pstmt.close();
            DBConnectionPool.returnDBConnection(dbconn, serialNumber);
        }//finally
    }

    private void getDocumentInfo(String docid) throws McdbException,
            AccessionNumberException, Exception
    {
        getDocumentInfo(new DocumentIdentifier(docid));
    }

    /**
     * Look up the document type information from the database
     *
     * @param docid
     *            the id of the document to look up
     */
    private void getDocumentInfo(DocumentIdentifier docid)
            throws McdbException, Exception
    {
        DBConnection dbconn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;
        String table = "xml_documents";

        try {
            if (isRevisionOnly(docid)) { //pull the document from xml_revisions
                                         // instead of from xml_documents;
                table = "xml_revisions";
            }
        }
        // catch a McdbDocNotFoundException throw it
        catch (McdbDocNotFoundException notFound) {
            throw notFound;
        } catch (Exception e) {

            MetaCatUtil.debugMessage("error in DocumentImpl.getDocumentInfo: "
                    + e.getMessage(), 30);
            throw e;
        }

        try {
            dbconn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getDocumentInfo");
            serialNumber = dbconn.getCheckOutSerialNumber();
            StringBuffer sql = new StringBuffer();
            sql.append("SELECT docname, doctype, rootnodeid, ");
            sql.append("date_created, date_updated, user_owner, user_updated,");
            sql.append(" server_location, public_access, rev");
            sql.append(" FROM ").append(table);
            sql.append(" WHERE docid LIKE '").append(docid.getIdentifier());
            sql.append("' and rev like '").append(docid.getRev()).append("'");

            pstmt = dbconn.prepareStatement(sql.toString());

            pstmt.execute();
            ResultSet rs = pstmt.getResultSet();
            boolean tableHasRows = rs.next();
            if (tableHasRows) {
                this.docname = rs.getString(1);
                this.doctype = rs.getString(2);
                this.rootnodeid = rs.getLong(3);
                this.createdate = rs.getString(4);
                this.updatedate = rs.getString(5);
                this.userowner = rs.getString(6);
                this.userupdated = rs.getString(7);
                this.serverlocation = rs.getInt(8);
                this.publicaccess = rs.getString(9);
                this.rev = rs.getInt(10);
            }
            pstmt.close();

            //get doc home server name
            pstmt = dbconn.prepareStatement("select server "
                    + "from xml_replication where serverid = ?");
            //because connection use twice here, so we need to increase one
            dbconn.increaseUsageCount(1);
            pstmt.setInt(1, serverlocation);
            pstmt.execute();
            rs = pstmt.getResultSet();
            tableHasRows = rs.next();
            if (tableHasRows) {

                String server = rs.getString(1);
                //get homeserver name
                if (!server.equals("localhost")) {
                    this.docHomeServer = server;
                } else {
                    this.docHomeServer = MetaCatUtil
                            .getLocalReplicationServerName();
                }
                MetaCatUtil.debugMessage("server: " + docHomeServer, 50);

            }
            pstmt.close();
            if (this.doctype != null) {
                pstmt = dbconn.prepareStatement("SELECT system_id, entry_type "
                        + "FROM xml_catalog " + "WHERE public_id = ?");
                //should increase usage count again
                dbconn.increaseUsageCount(1);
                // Bind the values to the query
                pstmt.setString(1, doctype);

                pstmt.execute();
                rs = pstmt.getResultSet();
                tableHasRows = rs.next();
                if (tableHasRows) {
                    this.system_id = rs.getString(1);
                    this.validateType = rs.getString(2);

                }
                pstmt.close();
            }
        } catch (SQLException e) {
            System.out.println("error in DocumentImpl.getDocumentInfo: "
                    + e.getMessage());
            e.printStackTrace(System.out);
            throw new McdbException("Error accessing database connection in "
                    + "DocumentImpl.getDocumentInfo: ", e);
        } finally {
            try {
                pstmt.close();
            } catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "error in DocumentImple.getDocumentInfo: "
                                + ee.getMessage(), 30);
            } finally {
                DBConnectionPool.returnDBConnection(dbconn, serialNumber);
            }
        }

        if (this.docname == null) {
            throw new McdbDocNotFoundException(
                "Document not found: " + docid, docid.getIdentifier(), docid
                        .getRev());
        }
    }

    /**
     * Look up the node data from the database, but some node would be shown
     * because of access control
     *
     * @param rootnodeid
     *            the id of the root node of the node tree to look up
     * @param accessControl
     *            the hashtable has control info
     */
    private TreeSet getPartNodeRecordList(long rootnodeid,
            Hashtable accessControl) throws McdbException
    {
        PreparedStatement pstmt = null;
        DBConnection dbconn = null;
        int serialNumber = -1;
        TreeSet nodeRecordList = new TreeSet(new NodeComparator());
        long nodeid = 0;
        long parentnodeid = 0;
        long nodeindex = 0;
        String nodetype = null;
        String nodename = null;
        String nodeprefix = null;
        String nodedata = null;
        String quotechar = dbAdapter.getStringDelimiter();
        String sql = "SELECT nodeid,parentnodeid,nodeindex, "
                + "nodetype,nodename,nodeprefix,nodedata "
                + "FROM xml_nodes WHERE rootnodeid = ?";

        // go through the access control for some nodes
        Enumeration en = accessControl.elements();
        while (en.hasMoreElements()) {
            SubTree tree = (SubTree) en.nextElement();
            long startId = tree.getStartNodeId();
            long endId = tree.getEndNodeId();
            sql = sql + " AND(nodeid < " + startId + " OR nodeid > " + endId
                    + ")";

        }
        MetaCatUtil.debugMessage("The final query to select part node tree: "
                + sql, 25);

        try {
            dbconn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getPartNodeRecordList");
            serialNumber = dbconn.getCheckOutSerialNumber();
            pstmt = dbconn.prepareStatement(sql);

            // Bind the values to the query
            pstmt.setLong(1, rootnodeid);
            pstmt.execute();
            ResultSet rs = pstmt.getResultSet();
            boolean tableHasRows = rs.next();
            while (tableHasRows) {
                nodeid = rs.getLong(1);
                parentnodeid = rs.getLong(2);
                nodeindex = rs.getLong(3);
                nodetype = rs.getString(4);
                nodename = rs.getString(5);
                nodeprefix = rs.getString(6);
                nodedata = rs.getString(7);
                nodedata = MetaCatUtil.normalize(nodedata);
                // add the data to the node record list hashtable
                NodeRecord currentRecord = new NodeRecord(nodeid, parentnodeid,
                        nodeindex, nodetype, nodename, nodeprefix, nodedata);
                nodeRecordList.add(currentRecord);

                // Advance to the next node
                tableHasRows = rs.next();
            }
            pstmt.close();

        } catch (SQLException e) {
            throw new McdbException(
                    "Error in DocumentImpl.getPartNodeRecordList "
                            + e.getMessage());
        } finally {
            try {
                pstmt.close();
            } catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "error in DocumentImpl.getPartNodeRecordList: "
                                + ee.getMessage(), 30);
            } finally {
                DBConnectionPool.returnDBConnection(dbconn, serialNumber);
            }
        }

        if (!nodeRecordList.isEmpty()) {

            return nodeRecordList;
        } else {

            throw new McdbException("Error getting node data: " + docid);
        }
    }

    /**
     * Look up the node data from the database
     *
     * @param rootnodeid
     *            the id of the root node of the node tree to look up
     */
    private TreeSet getNodeRecordList(long rootnodeid) throws McdbException
    {
        PreparedStatement pstmt = null;
        DBConnection dbconn = null;
        int serialNumber = -1;
        TreeSet nodeRecordList = new TreeSet(new NodeComparator());
        long nodeid = 0;
        long parentnodeid = 0;
        long nodeindex = 0;
        String nodetype = null;
        String nodename = null;
        String nodeprefix = null;
        String nodedata = null;
        String quotechar = dbAdapter.getStringDelimiter();

        try {
            dbconn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getNodeRecordList");
            serialNumber = dbconn.getCheckOutSerialNumber();
            pstmt = dbconn
                    .prepareStatement("SELECT nodeid,parentnodeid,nodeindex, "
                            + "nodetype,nodename,nodeprefix,nodedata "
                            + "FROM xml_nodes WHERE rootnodeid = ?");

            // Bind the values to the query
            pstmt.setLong(1, rootnodeid);

            pstmt.execute();
            ResultSet rs = pstmt.getResultSet();
            boolean tableHasRows = rs.next();
            while (tableHasRows) {
                nodeid = rs.getLong(1);
                parentnodeid = rs.getLong(2);
                nodeindex = rs.getLong(3);
                nodetype = rs.getString(4);
                nodename = rs.getString(5);
                nodeprefix = rs.getString(6);
                nodedata = rs.getString(7);
		nodedata = MetaCatUtil.normalize(nodedata);
                // add the data to the node record list hashtable
                NodeRecord currentRecord = new NodeRecord(nodeid, parentnodeid,
                        nodeindex, nodetype, nodename, nodeprefix, nodedata);
                nodeRecordList.add(currentRecord);

                // Advance to the next node
                tableHasRows = rs.next();
            }
            pstmt.close();

        } catch (SQLException e) {
            throw new McdbException("Error in DocumentImpl.getNodeRecordList "
                    + e.getMessage());
        } finally {
            try {
                pstmt.close();
            } catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "error in DocumentImpl.getNodeRecordList: "
                                + ee.getMessage(), 30);
            } finally {
                DBConnectionPool.returnDBConnection(dbconn, serialNumber);
            }
        }

        return nodeRecordList;

    }

    /** creates SQL code and inserts new document into DB connection */
    private void writeDocumentToDB(String action, String user, String pub,
            String catalogid, int serverCode) throws SQLException, Exception
    {
        String sysdate = dbAdapter.getDateTimeFunction();

        try {
            PreparedStatement pstmt = null;

            if (action.equals("INSERT")) {
                //AccessionNumber ac = new AccessionNumber();
                //this.docid = ac.generate(docid, "INSERT");

                pstmt = connection
                        .prepareStatement("INSERT INTO xml_documents "
                        + "(docid, rootnodeid, docname, doctype, user_owner, "
                        + "user_updated, date_created, date_updated, "
                        + "public_access, catalog_id, server_location, rev) "
                        + "VALUES (?, ?, ?, ?, ?, ?, " + sysdate + ", "
                        + sysdate + ", ?, ?, ?, ?)");
                // Increase dbconnection usage count
                connection.increaseUsageCount(1);

                //note that the server_location is set to 1.
                //this means that "localhost" in the xml_replication table must
                //always be the first entry!!!!!

                // Bind the values to the query
                pstmt.setString(1, this.docid);
                pstmt.setLong(2, rootnodeid);
                pstmt.setString(3, docname);
                pstmt.setString(4, doctype);
                pstmt.setString(5, user);
                pstmt.setString(6, user);
                //public access is usefulless, so set it to null
                pstmt.setString(7, null);
                /*
                 * if ( pub == null ) { pstmt.setString(7, null); } else if (
                 * pub.toUpperCase().equals("YES") || pub.equals("1") ) {
                 * pstmt.setInt(7, 1); } else if (
                 * pub.toUpperCase().equals("NO") || pub.equals("0") ) {
                 * pstmt.setInt(7, 0); }
                 */
                pstmt.setString(8, catalogid);
                pstmt.setInt(9, serverCode);
                pstmt.setInt(10, Integer.parseInt(updatedVersion));
            } else if (action.equals("UPDATE")) {

                // Save the old document publicaccessentry in a backup table
                DocumentImpl.archiveDocRevision(connection, docid, user);
                MetaCatUtil.debugMessage("after archiveDoc", 40);
                DocumentImpl thisdoc = new DocumentImpl(docid, false);
                int thisrev = thisdoc.getRev();
                MetaCatUtil.debugMessage("this revsion is: " + thisrev, 40);
                //if the updated vesion is not greater than current one,
                //throw it into a exception
                if (Integer.parseInt(updatedVersion) <= thisrev) {
                    throw new Exception("Next revision number couldn't be less"
                            + " than or equal " + thisrev);
                } else {
                    //set the user specified revision
                    thisrev = Integer.parseInt(updatedVersion);
                }
                MetaCatUtil.debugMessage("final revsion is: " + thisrev, 40);
                boolean useXMLIndex = (new Boolean(MetaCatUtil
                        .getOption("usexmlindex"))).booleanValue();
                if (useXMLIndex) {
                    MetaCatUtil.debugMessage("before delete", 40);
                    // Delete index for the old version of docid
                    // The new index is inserting on the next calls to DBSAXNode
                    pstmt = connection
                            .prepareStatement("DELETE FROM xml_index WHERE docid='"
                                    + this.docid + "'");
                    MetaCatUtil.debugMessage("after delete", 40);
                    // Increase dbconnection usage count
                    connection.increaseUsageCount(1);

                    pstmt.execute();
                    pstmt.close();
                }

                // Update the new document to reflect the new node tree
                pstmt = connection
                        .prepareStatement("UPDATE xml_documents "
                        + "SET rootnodeid = ?, docname = ?, doctype = ?, "
                        + "user_updated = ?, date_updated = "
                        + sysdate
                        + ", "
                        + "server_location = ?, rev = ?, public_access = ?, "
                        + "catalog_id = ? "
                        + "WHERE docid = ?");
                // Increase dbconnection usage count
                connection.increaseUsageCount(1);
                // Bind the values to the query
                pstmt.setLong(1, rootnodeid);
                pstmt.setString(2, docname);
                pstmt.setString(3, doctype);
                pstmt.setString(4, user);
                pstmt.setInt(5, serverCode);
                pstmt.setInt(6, thisrev);
                pstmt.setString(7, null);
                /*
                 * if ( pub == null ) { pstmt.setString(7, null); } else if (
                 * pub.toUpperCase().equals("YES") || pub.equals("1") ) { pstmt
                 * .setInt(7, 1); } else if ( pub.toUpperCase().equals("NO") ||
                 * pub.equals("0") ) { pstmt.setInt(7, 0); }
                 */
                pstmt.setString(8, catalogid);
                pstmt.setString(9, this.docid);

            } else {
                System.err.println("Action not supported: " + action);
            }

            // Do the insertion
            pstmt.execute();

            pstmt.close();

        } catch (SQLException sqle) {
            throw sqle;
        } catch (Exception e) {
            throw e;
        }
    }

    /**
     * Write an XML file to the database, given a filename
     *
     * @param conn
     *            the JDBC connection to the database
     * @param filename
     *            the filename to be loaded into the database
     * @param pub
     *            flag for public "read" access on document
     * @param dtdfilename
     *            the dtd to be uploaded on server's file system
     * @param action
     *            the action to be performed (INSERT OR UPDATE)
     * @param docid
     *            the docid to use for the INSERT OR UPDATE
     * @param user
     *            the user that owns the document
     * @param groups
     *            the groups to which user belongs
     */
    /*
     * public static String write(DBConnection conn,String filename, String pub,
     * String dtdfilename, String action, String docid, String user, String[]
     * groups ) throws Exception {
     *
     * Reader dtd = null; if ( dtdfilename != null ) { dtd = new FileReader(new
     * File(dtdfilename).toString()); } return write ( conn, new FileReader(new
     * File(filename).toString()), pub, dtd, action, docid, user, groups,
     * false); }
     */

    public static String write(DBConnection conn, Reader xml, String pub,
            Reader dtd, String action, String docid, String user,
            String[] groups, String ruleBase, boolean needValidation)
            throws Exception
    {
        //this method will be called in handleUpdateOrInsert method
        //in MetacatServlet class and now is wrapper into documentImple
        // get server location for this doc
        int serverLocation = getServerLocationNumber(docid);
        return write(conn, xml, pub, dtd, action, docid, user, groups,
                serverLocation, false, ruleBase, needValidation);
    }

    /**
     * Write an XML file to the database, given a Reader
     *
     * @param conn
     *            the JDBC connection to the database
     * @param xml
     *            the xml stream to be loaded into the database
     * @param pub
     *            flag for public "read" access on xml document
     * @param dtd
     *            the dtd to be uploaded on server's file system
     * @param action
     *            the action to be performed (INSERT or UPDATE)
     * @param accnum
     *            the docid + rev# to use on INSERT or UPDATE
     * @param user
     *            the user that owns the document
     * @param groups
     *            the groups to which user belongs
     * @param serverCode
     *            the serverid from xml_replication on which this document
     *            resides.
     * @param override
     *            flag to stop insert replication checking. if override = true
     *            then a document not belonging to the local server will not be
     *            checked upon update for a file lock. if override = false then
     *            a document not from this server, upon update will be locked
     *            and version checked.
     */

    public static String write(DBConnection conn, Reader xml, String pub,
            Reader dtd, String action, String accnum, String user,
            String[] groups, int serverCode, boolean override, String ruleBase,
            boolean needValidation) throws Exception
    {
        // NEW - WHEN CLIENT ALWAYS PROVIDE ACCESSION NUMBER INCLUDING REV IN IT
        //MetaCatUtil util = new MetaCatUtil();
        MetaCatUtil.debugMessage("conn usage count before writting: "
                + conn.getUsageCount(), 50);
        AccessionNumber ac = new AccessionNumber(accnum, action);
        String docid = ac.getDocid();
        String rev = ac.getRev();
        MetaCatUtil.debugMessage("action: " + action + " servercode: "
                + serverCode + " override: " + override, 10);

        if ((serverCode != 1 && action.equals("UPDATE")) && !override) {
            // if this document being written is not a resident of this server
            // then we need to try to get a lock from it's resident server. If
            // the resident server will not give a lock then we send the user
            // a  message saying that he/she needs to download a new copy of
            // the file and merge the differences manually.
            int istreamInt;
            char istreamChar;

            // check for 'write' permission for 'user' to update this document
            if (!hasWritePermission(user, groups, accnum)) {
                throw new Exception(
                    "User " + user
                    + " does not have permission to update XML Document #"
                    + accnum);
            }

            DocumentIdentifier id = new DocumentIdentifier(accnum);
            String updaterev = id.getRev();
            String server = MetacatReplication
                    .getServerNameForServerCode(serverCode);
            MetacatReplication.replLog("attempting to lock " + accnum);
            URL u = new URL("https://" + server + "?server="
                    + MetaCatUtil.getLocalReplicationServerName()
                    + "&action=getlock&updaterev=" + updaterev + "&docid="
                    + docid);
            //System.out.println("sending message: " + u.toString());
            String serverResStr = MetacatReplication.getURLContent(u);
            String openingtag = serverResStr.substring(0, serverResStr
                    .indexOf(">") + 1);
            if (openingtag.equals("<lockgranted>")) {//the lock was granted go
                                                     // ahead with the insert
                XMLReader parser = null;
                try {
                    //System.out.println("In lockgranted");
                    MetacatReplication.replLog("lock granted for " + accnum
                            + " from " + server);
                    /*
                     * XMLReader parser = initializeParser(conn, action, docid,
                     * updaterev, validate, user, groups, pub, serverCode, dtd);
                     */
                    parser = initializeParser(conn, action, docid, updaterev,
                            user, groups, pub, serverCode, dtd, ruleBase,
                            needValidation);
                    conn.setAutoCommit(false);
                    parser.parse(new InputSource(xml));
                    conn.commit();
                    conn.setAutoCommit(true);
                } catch (Exception e) {
                    conn.rollback();
                    conn.setAutoCommit(true);
                    //if it is a eml2 document, we need delete online data
                    if (parser != null) {
                        ContentHandler handler = parser.getContentHandler();
                        if (handler instanceof Eml200SAXHandler) {
                            Eml200SAXHandler eml = (Eml200SAXHandler) handler;
                            eml.deleteInlineFiles();
                        }
                    }
                    throw e;
                }
                // run write into access db base one relation table and access
                // object
                runRelationAndAccessHandler(accnum, user, groups, serverCode);

                // Force replication the docid
                ForceReplicationHandler frh = new ForceReplicationHandler(
                        accnum, true, null);
                return (accnum);

            }

            else if (openingtag.equals("<filelocked>")) {
                // the file is currently locked by another user notify our
                // user to wait a few minutes, check out a new copy and try
                // again.
                MetacatReplication.replLog("lock denied for " + accnum + " on "
                        + server + " reason: file already locked");
                throw new Exception(
                        "The file specified is already locked by another "
                                + "user.  Please wait 30 seconds, checkout the "
                                + "newer document, merge your changes and try "
                                + "again.");
            } else if (openingtag.equals("<outdatedfile>")) {
                // our file is outdated. notify our user to check out a new
                // copy of the file and merge his version with the new version.
                //System.out.println("outdated file");
                MetacatReplication.replLog("lock denied for " + accnum + " on "
                        + server + " reason: local file outdated");
                throw new Exception(
                        "The file you are trying to update is an outdated"
                        + " version.  Please checkout the newest document, "
                        + "merge your changes and try again.");
            }
        }

        if (action.equals("UPDATE")) {
            // check for 'write' permission for 'user' to update this document
            if (!hasWritePermission(user, groups, accnum)) {
                throw new Exception(
                    "User " + user
                    + " does not have permission to update XML Document #"
                    + accnum); }
        }
        XMLReader parser = null;
        try {
            parser = initializeParser(conn, action, docid, rev, user, groups,
                    pub, serverCode, dtd, ruleBase, needValidation);

            conn.setAutoCommit(false);
            parser.parse(new InputSource(xml));
            conn.commit();
            conn.setAutoCommit(true);
        } catch (Exception e) {
            conn.rollback();
            conn.setAutoCommit(true);
            //if it is a eml2 document, we need delete online data
            if (parser != null) {
                ContentHandler handler = parser.getContentHandler();
                if (handler instanceof Eml200SAXHandler) {
                    Eml200SAXHandler eml = (Eml200SAXHandler) handler;
                    eml.deleteInlineFiles();
                }
            }
            throw e;
        }

        // run access db base on relation table and access object
        runRelationAndAccessHandler(accnum, user, groups, serverCode);

        // Delete enteries from xml_queryresult for given docid if
        // action is UPDATE
        // These enteries will be created again when the docid is part of a
        // result next time
        if (action.equals("UPDATE")) {
          try {
              PreparedStatement pstmt = null;
              pstmt = conn.prepareStatement(
                      "DELETE FROM xml_queryresult WHERE docid = ?");
              pstmt.setString(1, docid);
              pstmt.execute();
              pstmt.close();
              conn.increaseUsageCount(1);
          } catch (Exception e){
              MetaCatUtil.debugMessage("Error in deleting enteries from "
                                       + "xml_queryresult where docid is "
                                       + docid + " in DBQuery.write: "
                                       + e.getMessage(), 20);
           }

        }

        // Force replicate out the new document to each server in our server
        // list. Start the thread to replicate this new document out to the
        // other servers true mean it is xml document null is because no
        // metacat notify the force replication.
        ForceReplicationHandler frh = new ForceReplicationHandler(accnum,
                action, true, null);

        MetaCatUtil.debugMessage("Conn Usage count after writting: "
                + conn.getUsageCount(), 50);
        return (accnum);
    }

    /**
     * Write an XML file to the database during replication
     *
     * @param conn
     *            the JDBC connection to the database
     * @param xml
     *            the xml stream to be loaded into the database
     * @param pub
     *            flag for public "read" access on xml document
     * @param dtd
     *            the dtd to be uploaded on server's file system
     * @param action
     *            the action to be performed (INSERT or UPDATE)
     * @param accnum
     *            the docid + rev# to use on INSERT or UPDATE
     * @param user
     *            the user that owns the document
     * @param groups
     *            the groups to which user belongs
     * @param homeServer
     *            the name of server which the document origanlly create
     * @param validate,
     *            if the xml document is valid or not
     * @param notifyServer,
     *            the server which notify local server the force replication
     *            command
     */
    public static String writeReplication(DBConnection conn, Reader xml,
            String pub, Reader dtd, String action, String accnum, String user,
            String[] groups, String homeServer, String notifyServer,
            String ruleBase, boolean needValidation) throws Exception
    {
        MetaCatUtil.debugMessage("user in replication" + user, 30);
        // Docid without revision
        String docid = MetaCatUtil.getDocIdFromAccessionNumber(accnum);
        // Revision specified by user (int)
        int userSpecifyRev = MetaCatUtil.getRevisionFromAccessionNumber(accnum);
        MetaCatUtil.debugMessage("The user specifyRev: " + userSpecifyRev, 30);
        // Revision for this docid in current database
        int revInDataBase = getLatestRevisionNumber(docid);
        MetaCatUtil.debugMessage("The rev in data base: " + revInDataBase, 30);
        // String to store the revision
        String rev = null;

        //revIndataBase=-1, there is no record in xml_documents table
        //the document is a new one for local server, inert it into table
        //user specified rev should be great than 0
        if (revInDataBase == -1 && userSpecifyRev >= 0) {
            // rev equals user specified
            rev = (new Integer(userSpecifyRev)).toString();
            // action should be INSERT
            action = "INSERT";
        }
        //rev is greater the last revsion number and revInDataBase isn't -1
        // it is a updated file
        else if (userSpecifyRev > revInDataBase && revInDataBase >= 0) {
            // rev equals user specified
            rev = (new Integer(userSpecifyRev)).toString();
            // action should be update
            action = "UPDATE";
        }
        // local server has newer version, then notify the remote server
        else if (userSpecifyRev < revInDataBase && revInDataBase > 0) {
            throw new Exception("Local server: "
                    + MetaCatUtil.getOption("server")
                    + " has newer revision of doc: " + docid + "."
                    + revInDataBase + ". Please notify it.");
        }
        //other situation
        else {

            throw new Exception("The docid" + docid
                    + "'s revision number couldn't be " + userSpecifyRev);
        }
        // Variable to store homeserver code
        int serverCode = -2;

        // If server is not int the xml replication talbe, insert it into
        // xml_replication table
        //serverList.addToServerListIfItIsNot(homeServer);
        insertServerIntoReplicationTable(homeServer);
        // Get server code again
        serverCode = getServerCode(homeServer);

        MetaCatUtil
                .debugMessage("Document " + docid + "." + rev + " " + action
                        + " into local" + " metacat with servercode: "
                        + serverCode, 10);

        // insert into xml_nodes table
        XMLReader parser = null;
        try {

            parser = initializeParser(conn, action, docid, rev, user, groups,
                    pub, serverCode, dtd, ruleBase, needValidation);
            conn.setAutoCommit(false);
            parser.parse(new InputSource(xml));
            conn.commit();
            conn.setAutoCommit(true);
        } catch (Exception e) {
            conn.rollback();
            conn.setAutoCommit(true);
            if (parser != null) {
                ContentHandler handler = parser.getContentHandler();
                if (handler instanceof Eml200SAXHandler) {
                    Eml200SAXHandler eml = (Eml200SAXHandler) handler;
                    eml.deleteInlineFiles();
                }
            }
            throw e;
        }

        // run write into access db base on relation table and access rule
        try {
            runRelationAndAccessHandler(accnum, user, groups, serverCode);
        } catch (Exception ee) {
            MetacatReplication.replErrorLog("Failed to " + "create access "
                    + "rule for package: " + accnum + " because "
                    + ee.getMessage());
            MetaCatUtil.debugMessage("Failed to  " + "create access "
                    + "rule for package: " + accnum + " because "
                    + ee.getMessage(), 30);
        }
        //Force replication to other server
        ForceReplicationHandler forceReplication = new ForceReplicationHandler(
                accnum, action, true, notifyServer);

        return (accnum);
    }

    /* Running write record to xml_relation and xml_access */
    private static void runRelationAndAccessHandler(String accnumber,
            String userName, String[] group, int servercode) throws Exception
    {
        DBConnection dbconn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;
        String documenttype = getDocTypeFromDBForCurrentDocument(accnumber);
        try {
            String packagedoctype = MetaCatUtil.getOption("packagedoctype");
            Vector packagedoctypes = new Vector();
            packagedoctypes = MetaCatUtil.getOptionList(packagedoctype);
            String docIdWithoutRev = MetaCatUtil.getDocIdFromString(accnumber);
            if (documenttype != null &&
                    packagedoctypes.contains(documenttype)) {
                dbconn = DBConnectionPool.getDBConnection(
                        "DocumentImpl.runRelationAndAccessHandeler");
                serialNumber = dbconn.getCheckOutSerialNumber();
                dbconn.setAutoCommit(false);
                // from the relations get the access file id for that package
                String aclid = RelationHandler.getAccessFileID(docIdWithoutRev);
                // if there are access file, write ACL for that package
                if (aclid != null) {
                    runAccessControlList(dbconn, aclid, userName, group,
                            servercode);
                }
                dbconn.commit();
                dbconn.setAutoCommit(true);
            }
            // if it is an access file
            else if (documenttype != null
                    && MetaCatUtil.getOptionList(
                            MetaCatUtil.getOption("accessdoctype")).contains(
                            documenttype)) {
                dbconn = DBConnectionPool.getDBConnection(
                        "DocumentImpl.runRelationAndAccessHandeler");
                serialNumber = dbconn.getCheckOutSerialNumber();
                dbconn.setAutoCommit(false);
                // write ACL for the package
                runAccessControlList(dbconn, docIdWithoutRev, userName, group,
                        servercode);
                dbconn.commit();
                dbconn.setAutoCommit(true);

            }

        } catch (Exception e) {
            if (dbconn != null) {
                dbconn.rollback();
                dbconn.setAutoCommit(true);
            }
            MetaCatUtil.debugMessage(
                    "Error in DocumentImple.runRelationAndAccessHandler "
                            + e.getMessage(), 30);
            throw e;
        } finally {
            if (dbconn != null) {
                DBConnectionPool.returnDBConnection(dbconn, serialNumber);
            }
        }
    }

    // It runs in xmlIndex thread. It writes ACL for a package.
    private static void runAccessControlList(DBConnection conn, String aclid,
            String users, String[] group, int servercode) throws Exception
    {
        // read the access file from xml_nodes
        // parse the access file and store the access info into xml_access
        AccessControlList aclobj = new AccessControlList(conn, aclid, users,
                group, servercode);

    }

    /* Method get document type from db */
    private static String getDocTypeFromDBForCurrentDocument(String accnumber)
            throws SQLException
    {
        String docoumentType = null;
        String docid = null;
        PreparedStatement pstate = null;
        ResultSet rs = null;
        String sql = "SELECT doctype FROM xml_documents where docid = ?";
        DBConnection dbConnection = null;
        int serialNumber = -1;
        try {
            //get rid of revision number
            docid = MetaCatUtil.getDocIdFromString(accnumber);
            dbConnection = DBConnectionPool.getDBConnection(
                    "DocumentImpl.getDocTypeFromDBForCurrentDoc");
            serialNumber = dbConnection.getCheckOutSerialNumber();
            pstate = dbConnection.prepareStatement(sql);
            //bind variable
            pstate.setString(1, docid);
            //excute query
            pstate.execute();
            //handle resultset
            rs = pstate.getResultSet();
            if (rs.next()) {
                docoumentType = rs.getString(1);
            }
            rs.close();
            pstate.close();
        }//try
        catch (SQLException e) {
            MetaCatUtil.debugMessage("error in DocumentImpl."
                    + "getDocTypeFromDBForCurrentDocument " + e.getMessage(),
                    30);
            throw e;
        }//catch
        finally {
            pstate.close();
            DBConnectionPool.returnDBConnection(dbConnection, serialNumber);
        }//
        MetaCatUtil.debugMessage("The current doctype from db is: "
                + docoumentType, 35);
        return docoumentType;
    }

    /**
     * Delete an XML file from the database (actually, just make it a revision
     * in the xml_revisions table)
     *
     * @param docid
     *            the ID of the document to be deleted from the database
     */
    public static void delete(String accnum, String user, String[] groups, String notifyServer)
            throws Exception
    {

        DBConnection conn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;
        boolean isXML   = true;
        try {
            //check out DBConnection
            conn = DBConnectionPool.getDBConnection("DocumentImpl.delete");
            serialNumber = conn.getCheckOutSerialNumber();

            // NEW - WHEN CLIENT ALWAYS PROVIDE ACCESSION NUMBER INCLUDING REV
            // IN IT
            //AccessionNumber ac = new AccessionNumber(accnum, "DELETE");
            String docid = MetaCatUtil.getDocIdFromAccessionNumber(accnum);
            //String rev = ac.getRev();

            // Check if the document exists.
            pstmt = conn.prepareStatement("SELECT * FROM xml_documents "
                                          + "WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            ResultSet rs = pstmt.getResultSet();
            if(!rs.next()){
                rs.close();
                throw new Exception("Docid " + accnum + " does not exsist. "
                                    + "Please check that you have specified the"
                                    + " revision number of the document also.");
            }
            rs.close();
            pstmt.close();
            conn.increaseUsageCount(1);

            // get the type of deleting docid, this will be used in forcereplication
            String type = getDocTypeFromDB(conn, docid);
            if (type != null && type.trim().equals("BIN"))
            {
              isXML = false;
            }

            MetaCatUtil.debugMessage("Start deleting doc " + docid + "...", 20);
            // check for 'write' permission for 'user' to delete this document
            if (!hasAllPermission(user, groups, accnum)) {
                boolean adminIsAuthenticated = false;
                String[] administrators = null;
                String adminList = MetaCatUtil.getOption("administrators");
                try {
                    administrators = adminList.split(":");
                } catch (Exception pse) {
                    administrators = null;
                    MetaCatUtil.debugMessage("Error in DocumentImpl.delete: "
                                             + pse.getMessage(), 20);
                }
                for (int i = 0; i < administrators.length; i++) {
                    if (user.equals(administrators[i])) {
                        adminIsAuthenticated = true;
                    }
                }

                if(!adminIsAuthenticated){
                    throw new Exception(
                        "User " + user
                        + " does not have permission to delete XML Document #"
                        + accnum);
                }
            }

            conn.setAutoCommit(false);
            // Copy the record to the xml_revisions table
            DocumentImpl.archiveDocRevision(conn, docid, user);

            // Now delete it from the xml_index table
            boolean useXMLIndex = (new Boolean(MetaCatUtil
                    .getOption("usexmlindex"))).booleanValue();

           //if (useXMLIndex) {
            pstmt = conn
                    .prepareStatement("DELETE FROM xml_index WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            conn.increaseUsageCount(1);
            //}

            // Now delete it from xml_access table
            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_access WHERE accessfileid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            conn.increaseUsageCount(1);

            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_access WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            conn.increaseUsageCount(1);

            // Delete enteries from xml_queryresult
            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_queryresult WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            conn.increaseUsageCount(1);

            // Delete it from relation table
            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_relation WHERE docid = ?");
            //increase usage count
            conn.increaseUsageCount(1);
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();

            // Delete it from xml_accesssubtree table
            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_accesssubtree WHERE docid = ?");
            //increase usage count
            conn.increaseUsageCount(1);
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();

            // Delete it from xml_documents table
            pstmt = conn.prepareStatement(
                    "DELETE FROM xml_documents WHERE docid = ?");
            pstmt.setString(1, docid);
            pstmt.execute();
            pstmt.close();
            //Usaga count increase 1
            conn.increaseUsageCount(1);

            conn.commit();
            conn.setAutoCommit(true);

        } catch (Exception e) {
            MetaCatUtil.debugMessage("error in DocumentImpl.delete: "
                    + e.getMessage(), 30);
            throw e;
        } finally {

            try {
                // close preparedStatement
                if (pstmt != null) {
                    pstmt.close();
                }
            }
            finally {
                //check in DBonnection
                DBConnectionPool.returnDBConnection(conn, serialNumber);
            }
        }
       // add force delete replcation document here.
       ForceReplicationHandler frh = new ForceReplicationHandler(
                        accnum, ForceReplicationHandler.DELETE, isXML, notifyServer);
    }

    private static String getDocTypeFromDB(DBConnection conn, String docidWithoutRev)
                                 throws SQLException
    {
      String type = null;
      String sql = "SELECT DOCTYPE FROM xml_documents WHERE docid LIKE " + "'" +
                    docidWithoutRev +"'";
      Statement stmt = null;
      stmt = conn.createStatement();
      ResultSet result = stmt.executeQuery(sql);
      boolean hasResult = result.next();
      if (hasResult)
      {
        type = result.getString(1);
      }
      MetaCatUtil.debugMessage("The type of deleting docid " + docidWithoutRev +
                               " is " + type, 2);
      return type;
    }

    /**
     * Check for "WRITE" permission on @docid for @user and/or @groups
     * from DB connection
     */
    private static boolean hasWritePermission(String user, String[] groups,
            String docid) throws SQLException, Exception
    {
        // Check for WRITE permission on @docid for @user and/or @groups
        PermissionController controller = new PermissionController(docid);
        return controller.hasPermission(user, groups,
                AccessControlInterface.WRITESTRING);
    }

    /**
     * Check for "READ" permission base on docid, user and group
     *
     * @param docid, the document
     * @param user, user name
     * @param groups, user's group
     */
    public static boolean hasReadPermission(String user, String[] groups,
            String docId) throws SQLException, Exception
    {
        // Check for READ permission on @docid for @user and/or @groups
        PermissionController controller = new PermissionController(docId);
        return controller.hasPermission(user, groups,
                AccessControlInterface.READSTRING);
    }

    /**
     * Check for "WRITE" permission on @docid for @user and/or @groups
     * from DB connection
     */
    private static boolean hasAllPermission(String user, String[] groups,
            String docid) throws SQLException, Exception
    {
        // Check for WRITE permission on @docid for @user and/or @groups
        PermissionController controller = new PermissionController(docid);
        return controller.hasPermission(user, groups,
                AccessControlInterface.ALLSTRING);
    }

    /**
     * Set up the parser handlers for writing the document to the database
     */
    private static XMLReader initializeParser(DBConnection dbconn,
            String action, String docid, String rev, String user,
            String[] groups, String pub, int serverCode, Reader dtd,
            String ruleBase, boolean needValidation) throws Exception
    {
        XMLReader parser = null;
        try {
            // handler
            ContentHandler chandler;
            EntityResolver eresolver;
            DTDHandler dtdhandler;
            // Get an instance of the parser
            String parserName = MetaCatUtil.getOption("saxparser");
            parser = XMLReaderFactory.createXMLReader(parserName);
            if (ruleBase != null && ruleBase.equals(EML200)) {
                MetaCatUtil.debugMessage("eml 2.0.0 parser", 20);
                chandler = new Eml200SAXHandler(dbconn, action, docid, rev,
                        user, groups, pub, serverCode);
                parser.setContentHandler((ContentHandler) chandler);
                parser.setErrorHandler((ErrorHandler) chandler);
                parser.setProperty(DECLARATIONHANDLERPROPERTY, chandler);
                parser.setProperty(LEXICALPROPERTY, chandler);
                // turn on schema validation feature
                parser.setFeature(VALIDATIONFEATURE, true);
                parser.setFeature(NAMESPACEFEATURE, true);
                //parser.setFeature(NAMESPACEPREFIXESFEATURE, true);
                parser.setFeature(SCHEMAVALIDATIONFEATURE, true);
                // From DB to find the register external schema location
                String externalSchemaLocation = null;
                SchemaLocationResolver resolver = new SchemaLocationResolver();
                externalSchemaLocation = resolver
                        .getNameSpaceAndLocationString();
                // Set external schemalocation.
                if (externalSchemaLocation != null
                        && !(externalSchemaLocation.trim()).equals("")) {
                    parser.setProperty(EXTERNALSCHEMALOCATIONPROPERTY,
                            externalSchemaLocation);
                }
            } else if (ruleBase != null && ruleBase.equals(EML210)) {
                MetaCatUtil.debugMessage("eml 2.1.0 parser", 20);
                chandler = new Eml210SAXHandler(dbconn, action, docid, rev,
                        user, groups, pub, serverCode);
                parser.setContentHandler((ContentHandler) chandler);
                parser.setErrorHandler((ErrorHandler) chandler);
                parser.setProperty(DECLARATIONHANDLERPROPERTY, chandler);
                parser.setProperty(LEXICALPROPERTY, chandler);
                // turn on schema validation feature
                parser.setFeature(VALIDATIONFEATURE, true);
                parser.setFeature(NAMESPACEFEATURE, true);
                //parser.setFeature(NAMESPACEPREFIXESFEATURE, true);
                parser.setFeature(SCHEMAVALIDATIONFEATURE, true);
                // From DB to find the register external schema location
                String externalSchemaLocation = null;
                SchemaLocationResolver resolver = new SchemaLocationResolver();
                externalSchemaLocation = resolver
                        .getNameSpaceAndLocationString();
                // Set external schemalocation.
                if (externalSchemaLocation != null
                        && !(externalSchemaLocation.trim()).equals("")) {
                    parser.setProperty(EXTERNALSCHEMALOCATIONPROPERTY,
                            externalSchemaLocation);
                }
            } else {
                //create a DBSAXHandler object which has the revision
                // specification
                chandler = new DBSAXHandler(dbconn, action, docid, rev, user,
                        groups, pub, serverCode);
                parser.setContentHandler((ContentHandler) chandler);
                parser.setErrorHandler((ErrorHandler) chandler);
                parser.setProperty(DECLARATIONHANDLERPROPERTY, chandler);
                parser.setProperty(LEXICALPROPERTY, chandler);

                if (ruleBase != null && ruleBase.equals(SCHEMA)
                        && needValidation) {
                    MetaCatUtil.debugMessage("General schema parser", 20);
                    // turn on schema validation feature
                    parser.setFeature(VALIDATIONFEATURE, true);
                    parser.setFeature(NAMESPACEFEATURE, true);
                    //parser.setFeature(NAMESPACEPREFIXESFEATURE, true);
                    parser.setFeature(SCHEMAVALIDATIONFEATURE, true);
                    // From DB to find the register external schema location
                    String externalSchemaLocation = null;
                    SchemaLocationResolver resolver =
                        new SchemaLocationResolver();
                    externalSchemaLocation = resolver
                            .getNameSpaceAndLocationString();
                    // Set external schemalocation.
                    if (externalSchemaLocation != null
                            && !(externalSchemaLocation.trim()).equals("")) {
                        parser.setProperty(EXTERNALSCHEMALOCATIONPROPERTY,
                                externalSchemaLocation);
                    }

                } else if (ruleBase != null && ruleBase.equals(DTD)
                        && needValidation) {
                    MetaCatUtil.debugMessage("dtd parser", 20);
                    // turn on dtd validaton feature
                    parser.setFeature(VALIDATIONFEATURE, true);
                    eresolver = new DBEntityResolver(dbconn,
                            (DBSAXHandler) chandler, dtd);
                    dtdhandler = new DBDTDHandler(dbconn);
                    parser.setEntityResolver((EntityResolver) eresolver);
                    parser.setDTDHandler((DTDHandler) dtdhandler);
                } else {
                    MetaCatUtil.debugMessage("other parser", 20);
                    // non validation
                    parser.setFeature(VALIDATIONFEATURE, false);
                    eresolver = new DBEntityResolver(dbconn,
                            (DBSAXHandler) chandler, dtd);
                    dtdhandler = new DBDTDHandler(dbconn);
                    parser.setEntityResolver((EntityResolver) eresolver);
                    parser.setDTDHandler((DTDHandler) dtdhandler);
                }
            }//else
        } catch (Exception e) {
            throw e;
        }
        return parser;
    }

    /**
     * Save a document entry in the xml_revisions table Connection use as a
     * paramter is in order to rollback feature
     */
    private static void archiveDocRevision(DBConnection dbconn, String docid,
            String user)
    {
        String sysdate = dbAdapter.getDateTimeFunction();
        //DBConnection conn = null;
        //int serialNumber = -1;
        PreparedStatement pstmt = null;

        // create a record in xml_revisions table
        // for that document as selected from xml_documents

        try {
            //check out DBConnection
            /*
             * conn=DBConnectionPool.
             * getDBConnection("DocumentImpl.archiveDocRevision");
             * serialNumber=conn.getCheckOutSerialNumber();
             */
            pstmt = dbconn.prepareStatement("INSERT INTO xml_revisions "
                    + "(docid, rootnodeid, docname, doctype, "
                    + "user_owner, user_updated, date_created, date_updated, "
                    + "server_location, rev, public_access, catalog_id) "
                    + "SELECT ?, rootnodeid, docname, doctype, "
                    + "user_owner, ?, " + sysdate + ", " + sysdate + ", "
                    + "server_location, rev, public_access, catalog_id "
                    + "FROM xml_documents " + "WHERE docid = ?");
            // Increase dbconnection usage count
            dbconn.increaseUsageCount(1);
            // Bind the values to the query and execute it
            pstmt.setString(1, docid);
            pstmt.setString(2, user);
            pstmt.setString(3, docid);
            pstmt.execute();
            pstmt.close();
        } catch (SQLException e) {
            MetaCatUtil.debugMessage(
                    "Error in DocumentImpl.archiveDocRevision : "
                            + e.getMessage(), 30);
        } finally {
            try {
                pstmt.close();
            } catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "Error in DocumnetImpl.archiveDocRevision: "
                                + ee.getMessage(), 50);
            }
        }
    }

    /** Save a document entry in the xml_revisions table */
    private static void archiveDocRevision(String docid, String user)
    {
        String sysdate = dbAdapter.getDateTimeFunction();
        DBConnection conn = null;
        int serialNumber = -1;
        PreparedStatement pstmt = null;

        // create a record in xml_revisions table
        // for that document as selected from xml_documents

        try {
            //check out DBConnection
            conn = DBConnectionPool
                    .getDBConnection("DocumentImpl.archiveDocRevision");
            serialNumber = conn.getCheckOutSerialNumber();
            pstmt = conn.prepareStatement("INSERT INTO xml_revisions "
                    + "(docid, rootnodeid, docname, doctype, "
                    + "user_owner, user_updated, date_created, date_updated, "
                    + "server_location, rev, public_access, catalog_id) "
                    + "SELECT ?, rootnodeid, docname, doctype, "
                    + "user_owner, ?, " + sysdate + ", " + sysdate + ", "
                    + "server_location, rev, public_access, catalog_id "
                    + "FROM xml_documents " + "WHERE docid = ?");
            // Bind the values to the query and execute it
            pstmt.setString(1, docid);
            pstmt.setString(2, user);
            pstmt.setString(3, docid);
            pstmt.execute();
            pstmt.close();
        } catch (SQLException e) {
            MetaCatUtil.debugMessage(
                    "Error in DocumentImpl.archiveDocRevision : "
                            + e.getMessage(), 30);
        } finally {
            try {
                pstmt.close();
            } catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "Error in DocumnetImpl.archiveDocRevision: "
                                + ee.getMessage(), 50);
            } finally {
                //check in DBConnection
                DBConnectionPool.returnDBConnection(conn, serialNumber);
            }
        }
    }

    /**
     * delete a entry in xml_table for given docid
     *
     * @param docId,
     *            the id of the document need to be delete
     */
    private static void deleteXMLDocuments(String docId) throws SQLException
    {
        DBConnection conn = null;
        int serialNumber = -1;
        PreparedStatement pStmt = null;
        try {
            //check out DBConnection
            conn = DBConnectionPool
                    .getDBConnection("DocumentImpl.deleteXMLDocuments");
            serialNumber = conn.getCheckOutSerialNumber();
            //delete a record
            pStmt = conn.prepareStatement(
                    "DELETE FROM xml_documents WHERE docid = '" + docId + "'");
            pStmt.execute();
        } finally {
            try {
                pStmt.close();
            } catch (SQLException e) {
                MetaCatUtil.debugMessage(
                        "error in DocumentImpl.deleteXMLDocuments: "
                                + e.getMessage(), 50);
            } finally {
                //return back DBconnection
                DBConnectionPool.returnDBConnection(conn, serialNumber);
            }
        }
    }

    /**
     * Get last revision number from database for a docid If couldn't find an
     * entry, -1 will return The return value is integer because we want compare
     * it to there new one
     *
     * @param docid
     *            <sitecode>. <uniqueid>part of Accession Number
     */
    public static int getLatestRevisionNumber(String docId) throws SQLException
    {
        int rev = 1;
        PreparedStatement pStmt = null;
        DBConnection dbConn = null;
        int serialNumber = -1;
        // get rid of rev
        docId = MetaCatUtil.getDocIdFromString(docId);
        try {
            //check out DBConnection
            dbConn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getLatestRevisionNumber");
            serialNumber = dbConn.getCheckOutSerialNumber();

            pStmt = dbConn
                    .prepareStatement("SELECT rev FROM xml_documents WHERE docid='"
                            + docId + "'");
            pStmt.execute();

            ResultSet rs = pStmt.getResultSet();
            boolean hasRow = rs.next();
            if (hasRow) {
                rev = rs.getInt(1);
                pStmt.close();
            } else {
                rev = -1;
                pStmt.close();
            }
        }//try
        finally {
            try {
                pStmt.close();
            } catch (Exception ee) {
                MetaCatUtil.debugMessage("Error in DocumentImpl."
                        + "getLatestRevisionNumber: " + ee.getMessage(), 50);
            } finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }
        }//finally

        return rev;
    }//getLatestRevisionNumber

    /**
     * Get server location form database for a accNum
     *
     * @param accum
     *            <sitecode>. <uniqueid>. <rev>
     */
    private static int getServerLocationNumber(String accNum)
            throws SQLException
    {
        //get rid of revNum part
        String docId = MetaCatUtil.getDocIdFromString(accNum);
        PreparedStatement pStmt = null;
        int serverLocation = 1;
        DBConnection conn = null;
        int serialNumber = -1;

        try {
            //check out DBConnection
            conn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getServerLocationNumber");
            serialNumber = conn.getCheckOutSerialNumber();

            pStmt = conn
                    .prepareStatement("SELECT server_location FROM xml_documents WHERE docid='"
                            + docId + "'");
            pStmt.execute();

            ResultSet rs = pStmt.getResultSet();
            boolean hasRow = rs.next();
            //if there is entry in xml_documents, get the serverlocation
            if (hasRow) {
                serverLocation = rs.getInt(1);
                pStmt.close();
            } else {
                //if htere is no entry in xml_documents, we consider it is new
                // document
                //the server location is local host and value is 1
                serverLocation = 1;
                pStmt.close();
            }
        }//try
        finally {
            try {
                pStmt.close();
            }//try
            catch (Exception ee) {
                MetaCatUtil.debugMessage(
                        "Error in DocumentImpl.getServerLocationNu(): "
                                + ee.getMessage(), 50);
            }//catch
            finally {
                DBConnectionPool.returnDBConnection(conn, serialNumber);
            }//finally
        }//finally

        return serverLocation;
    }

    /**
     * Given a server name, return its servercode in xml_replication table. If
     * no server is found, -1 will return
     *
     * @param serverName,
     */
    private static int getServerCode(String serverName)
    {
        PreparedStatement pStmt = null;
        int serverLocation = -2;
        DBConnection dbConn = null;
        int serialNumber = -1;
        //MetaCatUtil util = new MetaCatUtil();

        //we should consider about local host too
        if (serverName.equals(MetaCatUtil.getLocalReplicationServerName())) {
            serverLocation = 1;
            return serverLocation;
        }

        try {
            //check xml_replication table
            //dbConn=util.openDBConnection();
            //check out DBConnection
            dbConn = DBConnectionPool
                    .getDBConnection("DocumentImpl.getServerCode");
            serialNumber = dbConn.getCheckOutSerialNumber();
            pStmt = dbConn
                    .prepareStatement("SELECT serverid FROM xml_replication WHERE server='"
                            + serverName + "'");
            pStmt.execute();

            ResultSet rs = pStmt.getResultSet();
            boolean hasRow = rs.next();
            //if there is entry in xml_replication, get the serverid
            if (hasRow) {
                serverLocation = rs.getInt(1);
                pStmt.close();
            } else {
                // if htere is no entry in xml_replication, -1 will return
                serverLocation = -1;
                pStmt.close();
            }
        } catch (Exception e) {
            MetaCatUtil.debugMessage("Error in DocumentImpl.getServerCode(): "
                    + e.getMessage(), 30);
        } finally {
            try {
                pStmt.close();
            } catch (Exception ee) {
                MetaCatUtil.debugMessage(
                        "Error in DocumentImpl.getServerCode(): "
                                + ee.getMessage(), 50);
            } finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }
        }

        return serverLocation;
    }

    /**
     * Insert a server into xml_replcation table
     *
     * @param server,
     *            the name of server
     */
    private static synchronized void insertServerIntoReplicationTable(
            String server)
    {
        PreparedStatement pStmt = null;
        DBConnection dbConn = null;
        int serialNumber = -1;

        // Initial value for the server
        int replicate = 0;
        int dataReplicate = 0;
        int hub = 0;

        try {
            // Get DBConnection
            dbConn = DBConnectionPool
                    .getDBConnection("DocumentImpl.insertServIntoReplicationTable");
            serialNumber = dbConn.getCheckOutSerialNumber();

            // Compare the server to dabase
            pStmt = dbConn
                    .prepareStatement("SELECT serverid FROM xml_replication WHERE server='"
                            + server + "'");
            pStmt.execute();
            ResultSet rs = pStmt.getResultSet();
            boolean hasRow = rs.next();
            // Close preparedstatement and result set
            pStmt.close();
            rs.close();

            // If the server is not in the table, and server is not local host,
            // insert it
            if (!hasRow
                    && !server.equals(MetaCatUtil
                            .getLocalReplicationServerName())) {
                // Set auto commit false
                dbConn.setAutoCommit(false);
                /*
                 * pStmt = dbConn.prepareStatement("INSERT INTO xml_replication " +
                 * "(server, last_checked, replicate, datareplicate, hub) " +
                 * "VALUES ('" + server + "', to_date(" + "'01/01/00',
                 * 'MM/DD/YY'), '" + replicate +"', '"+dataReplicate+"','"+ hub +
                 * "')");
                 */
                pStmt = dbConn
                        .prepareStatement("INSERT INTO xml_replication "
                                + "(server, last_checked, replicate, datareplicate, hub) "
                                + "VALUES ('" + server + "', "
                                + dbAdapter.toDate("01/01/1980", "MM/DD/YYYY")
                                + ", '" + replicate + "', '" + dataReplicate
                                + "','" + hub + "')");

                pStmt.execute();
                dbConn.commit();
                // Increase usage number
                dbConn.increaseUsageCount(1);
                pStmt.close();

            }
        }//try
        catch (Exception e) {
            MetaCatUtil.debugMessage(
                    "Error in DocumentImpl.insertServerIntoRepli(): "
                            + e.getMessage(), 30);
        }//catch
        finally {

            try {
                // Set auto commit true
                dbConn.setAutoCommit(true);
                pStmt.close();

            }//try
            catch (Exception ee) {
                MetaCatUtil.debugMessage(
                        "Error in DocumentImpl.insetServerIntoRepl(): "
                                + ee.getMessage(), 50);
            }//catch
            finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }

        }//finally

    }

    /**
     * the main routine used to test the DBWriter utility.
     * <p>
     * Usage: java DocumentImpl <-f filename -a action -d docid>
     *
     * @param filename
     *            the filename to be loaded into the database
     * @param action
     *            the action to perform (READ, INSERT, UPDATE, DELETE)
     * @param docid
     *            the id of the document to process
     */
    static public void main(String[] args)
    {
        DBConnection dbconn = null;
        int serialNumber = -1;
        try {
            String filename = null;
            String dtdfilename = null;
            String action = null;
            String docid = null;
            boolean showRuntime = false;
            boolean useOldReadAlgorithm = false;

            // Parse the command line arguments
            for (int i = 0; i < args.length; ++i) {
                if (args[i].equals("-f")) {
                    filename = args[++i];
                } else if (args[i].equals("-r")) {
                    dtdfilename = args[++i];
                } else if (args[i].equals("-a")) {
                    action = args[++i];
                } else if (args[i].equals("-d")) {
                    docid = args[++i];
                } else if (args[i].equals("-t")) {
                    showRuntime = true;
                } else if (args[i].equals("-old")) {
                    useOldReadAlgorithm = true;
                } else {
                    System.err.println("   args[" + i + "] '" + args[i]
                            + "' ignored.");
                }
            }

            // Check if the required arguments are provided
            boolean argsAreValid = false;
            if (action != null) {
                if (action.equals("INSERT")) {
                    if (filename != null) {
                        argsAreValid = true;
                    }
                } else if (action.equals("UPDATE")) {
                    if ((filename != null) && (docid != null)) {
                        argsAreValid = true;
                    }
                } else if (action.equals("DELETE")) {
                    if (docid != null) {
                        argsAreValid = true;
                    }
                } else if (action.equals("READ")) {
                    if (docid != null) {
                        argsAreValid = true;
                    }
                }
            }

            // Print usage message if the arguments are not valid
            if (!argsAreValid) {
                System.err.println("Wrong number of arguments!!!");
                System.err
                        .println("USAGE: java DocumentImpl [-t] <-a INSERT> [-d docid] <-f filename> "
                                + "[-r dtdfilename]");
                System.err
                        .println("   OR: java DocumentImpl [-t] <-a UPDATE -d docid -f filename> "
                                + "[-r dtdfilename]");
                System.err
                        .println("   OR: java DocumentImpl [-t] <-a DELETE -d docid>");
                System.err
                        .println("   OR: java DocumentImpl [-t] [-old] <-a READ -d docid>");
                return;
            }

            // Time the request if asked for
            double startTime = System.currentTimeMillis();

            // Open a connection to the database
            MetaCatUtil util = new MetaCatUtil();

            dbconn = DBConnectionPool.getDBConnection("DocumentImpl.main");
            serialNumber = dbconn.getCheckOutSerialNumber();

            double connTime = System.currentTimeMillis();
            // Execute the action requested (READ, INSERT, UPDATE, DELETE)
            if (action.equals("READ")) {
                DocumentImpl xmldoc = new DocumentImpl(docid);
                if (useOldReadAlgorithm) {
                    System.out.println(xmldoc.readUsingSlowAlgorithm());
                } else {
                    xmldoc.toXml(new PrintWriter(System.out), null, null, true);
                }
            } else if (action.equals("DELETE")) {
                DocumentImpl.delete(docid, null, null,null);
                System.out.println("Document deleted: " + docid);
            } else {
                /*
                 * String newdocid = DocumentImpl.write(dbconn, filename, null,
                 * dtdfilename, action, docid, null, null); if ((docid != null) &&
                 * (!docid.equals(newdocid))) { if (action.equals("INSERT")) {
                 * System.out.println("New document ID generated!!! "); } else
                 * if (action.equals("UPDATE")) { System.out.println("ERROR:
                 * Couldn't update document!!! "); } } else if ((docid == null) &&
                 * (action.equals("UPDATE"))) { System.out.println("ERROR:
                 * Couldn't update document!!! "); }
                 * System.out.println("Document processing finished for: " +
                 * filename + " (" + newdocid + ")");
                 */
            }

            double stopTime = System.currentTimeMillis();
            double dbOpenTime = (connTime - startTime) / 1000;
            double insertTime = (stopTime - connTime) / 1000;
            double executionTime = (stopTime - startTime) / 1000;
            if (showRuntime) {
                System.out.println("\n\nTotal Execution time was: "
                        + executionTime + " seconds.");
                System.out.println("Time to open DB connection was: "
                        + dbOpenTime + " seconds.");
                System.out.println("Time to insert document was: " + insertTime
                        + " seconds.");
            }
            dbconn.close();
        } catch (McdbException me) {
            me.toXml(new PrintWriter(System.err));
        } catch (AccessionNumberException ane) {
            System.out.println(ane.getMessage());
        } catch (Exception e) {
            System.err.println("EXCEPTION HANDLING REQUIRED");
            System.err.println(e.getMessage());
            e.printStackTrace(System.err);
        } finally {
            // Return db connection
            DBConnectionPool.returnDBConnection(dbconn, serialNumber);
        }
    }
}
