/**
 *  '$RCSfile: MetaCatUtil.java,v $'
 *    Purpose: A Class that implements utility methods for a metadata catalog
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones, Jivka Bojilova
 *    Release: 1.5.0
 *
 *   '$Author: sgarg $'
 *     '$Date: 2005/04/04 23:45:17 $'
 * '$Revision: 1.48 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

import edu.ucsb.nceas.dbadapter.AbstractDatabase;
import edu.ucsb.nceas.utilities.Options;

/**
 * A suite of utility classes for the metadata catalog server
 */
public class MetaCatUtil
{

    public static AbstractDatabase dbAdapter;

    private static Options options = null;

    private static boolean debug = true;


    /**
     * Determine our db adapter class and create an instance of that class
     */
    static {
        try {
            dbAdapter = (AbstractDatabase) createObject(getOption("dbAdapter"));
        } catch (Exception e) {
            System.err.println("Error in MetaCatUtil static block:"
                    + e.getMessage());
        }
    }

    /**
     * Instantiate a class using the name of the class at runtime
     *
     * @param className the fully qualified name of the class to instantiate
     */
    public static Object createObject(String className) throws Exception
    {

        Object object = null;
        try {
            Class classDefinition = Class.forName(className);
            object = classDefinition.newInstance();
        } catch (InstantiationException e) {
            throw e;
        } catch (IllegalAccessException e) {
            throw e;
        } catch (ClassNotFoundException e) {
            throw e;
        }
        return object;
    }

    /**
     * Utility method to get an option value from the properties file
     *
     * @param optionName the name of the option requested
     * @return the String value for the option, or null if not set
     */
    public static String getOption(String optionName)
    {
        if (options == null) {
            options = Options.getInstance();
        }
        String value = options.getOption(optionName);
        return value;
    }

    /** Utility method to convert a file handle into a URL */
    public static URL fileToURL(File file)
    {
        String path = file.getAbsolutePath();
        String fSep = System.getProperty("file.separator");
        if (fSep != null && fSep.length() == 1)
                path = path.replace(fSep.charAt(0), '/');
        if (path.length() > 0 && path.charAt(0) != '/') path = '/' + path;
        try {
            return new URL("file", null, path);
        } catch (java.net.MalformedURLException e) {
            /*
             * According to the spec this could only happen if the file
             */
            throw new Error("unexpected MalformedURLException");
        }
    }

    /**
     * Utility method to parse the query part of a URL into parameters. This
     * method assumes the format of the query par tof the url is an ampersand
     * separated list of name/value pairs, with equal signs separating the name
     * from the value (e.g., name=tom&zip=99801 ). Returns a has of the name
     * value pairs, hashed on name.
     */
    public static Hashtable parseQuery(String query)
            throws MalformedURLException
    {
        String[][] params = new String[200][2];
        Hashtable parameters = new Hashtable();

        String temp = "";
        boolean ampflag = true;
        boolean poundflag = false;
        int arrcount = 0;

        if (query != null) {
            for (int i = 0; i < query.length(); i++) {

                // go throught the remainder of the query one character at a
                // time.
                if (query.charAt(i) == '=') {
                    // if the current char is a # then the preceding should be
                    // a name
                    if (!poundflag && ampflag) {
                        params[arrcount][0] = temp.trim();
                        temp = "";
                    } else {
                        //if there are two #s or &s in a row throw an
                        // exception.
                        throw new MalformedURLException(
                                "metacatURL: Two parameter names "
                                        + "not allowed in sequence");
                    }
                    poundflag = true;
                    ampflag = false;
                } else if (query.charAt(i) == '&' || i == query.length() - 1) {
                    //the text preceding the & should be the param value.
                    if (i == query.length() - 1) {
                        //if at the end of the string grab the last value and
                        // append it.
                        if (query.charAt(i) != '=') {
                            //ignore an extra & on the end of the string
                            temp += query.charAt(i);
                        }
                    }

                    if (!ampflag && poundflag) {
                        params[arrcount][1] = temp.trim();
                        parameters
                                .put(params[arrcount][0], params[arrcount][1]);
                        temp = "";
                        arrcount++; //increment the array to the next row.
                    } else {
                        //if there are two =s or &s in a row through an
                        // exception
                        throw new MalformedURLException(
                                "metacatURL: Two parameter values "
                                        + "not allowed in sequence");
                    }
                    poundflag = false;
                    ampflag = true;
                } else {
                    //get the next character in the string
                    temp += query.charAt(i);
                }
            }
        }
        return parameters;
    }

    /**
     * Utility method to print debugging messages. User can set debug level for
     * this message. The number is fewer, the message is more important
     *
     * @param msg, the content of the message
     * @param debugLevel, an integer indicating the message debug leve
     */
    public static void debugMessage(String msg, int debugLevel)
    {
        if (debug) {
            int limit = 1;
            try {
                limit = Integer.parseInt(getOption("debuglevel"));

            } catch (Exception e) {
                System.out.println(e.getMessage());
            }
            //don't allow the user set debugLevel less than or equals 0
            if (debugLevel <= 0) {
                debugLevel = 1;
            }

            if (debugLevel < limit) {
                System.err.println("MetaCat: " + msg);
            }
        }
    }

    public static Vector getOptionList(String optiontext)
    {
        Vector optionsVector = new Vector();
        if (optiontext.indexOf(",") == -1) {
            optionsVector.addElement(optiontext);
            return optionsVector;
        }

        while (optiontext.indexOf(",") != -1) {
            String s = optiontext.substring(0, optiontext.indexOf(","));
            optionsVector.addElement(s.trim());
            optiontext = optiontext.substring(optiontext.indexOf(",") + 1,
                    optiontext.length());
            if (optiontext.indexOf(",") == -1) { //catch the last list entry
                optionsVector.addElement(optiontext.trim());
            }
        }
        return optionsVector;
    }

    /** Normalizes the given string. Taken from configXML.java */
    public static String normalize(String s)
    {
        StringBuffer str = new StringBuffer();

             int len = (s != null) ? s.length() : 0;
             for (int i = 0; i < len; i++) {
                 char ch = s.charAt(i);
                 switch (ch) {
                     case '<': {
                         str.append("&lt;");
                         break;
                     }
                     case '>': {
                         str.append("&gt;");
                         break;
                     }
                     case '&': {
                         /*
                          * patch provided by Johnoel Ancheta from U of Hawaii
                          */
                         // check if & is for a character reference &#xnnnn;
                         if (i + 1 < len - 1 && s.charAt(i + 1) == '#') {
                             str.append("&#");
                             i += 2;

                             ch = s.charAt(i);
                             while (i < len && ch != ';') {
                                 str.append(ch);
                                 i++;
                                 ch = s.charAt(i);
                             }
                             str.append(';');
                         }
                         else
                             str.append("&amp;");
                         /////////
                         break;
                     }
                    default: {
                         if ( (ch<128) && (ch>31) ) {
                             str.append(ch);
                         }
                         else if (ch<32) {
                             if (ch == 10) { // new line
                                 str.append(ch);
                             }
                             if (ch == 13) { // carriage return
                                 str.append(ch);
                             }
                             if (ch == 9) {  // tab
                                 str.append(ch);
                             }
                             // otherwise skip
                         }
                         else {
                             str.append("&#");
                             str.append(Integer.toString(ch));
                             str.append(';');
                         }
                     }
                 }
             }
             return str.toString();
    }

    /**
     * Get docid from online/url string
     */
    public static String getDocIdWithRevFromOnlineURL(String url)
    {
        String docid = null;
        String DOCID = "docid";
        boolean find = false;
        char limited = '&';
        int count = 0; //keep track how many & was found
        Vector list = new Vector();// keep index number for &
        if (url == null) {
            MetaCatUtil.debugMessage("url is null and null will be returned",
                    30);
            return docid;
        }
        // the first element in list is 0
        list.add(new Integer(0));
        for (int i = 0; i < url.length(); i++) {
            if (url.charAt(i) == limited) {
                // count plus 1
                count++;
                list.add(new Integer(i));
                // get substring beween two &
                String str = url.substring(
                        ((Integer) list.elementAt(count - 1)).intValue(), i);
                MetaCatUtil.debugMessage("substring between two & is: " + str,
                        30);
                //if the subString contains docid, we got it
                if (str.indexOf(DOCID) != -1) {
                    //get index of '="
                    int start = getIndexForGivenChar(str, '=') + 1;
                    int end = str.length();
                    docid = str.substring(start, end);
                    find = true;
                }//if
            }//if
        }//for
        //if not find, we need check the subtring between the index of last &
        // and
        // the end of string
        if (!find) {
            MetaCatUtil.debugMessage("Checking the last substring", 35);
            String str = url.substring(((Integer) list.elementAt(count))
                    .intValue() + 1, url.length());
            MetaCatUtil.debugMessage("Last substring is: " + str, 30);
            if (str.indexOf(DOCID) != -1) {
                //get index of '="
                int start = getIndexForGivenChar(str, '=') + 1;
                int end = str.length();
                docid = str.substring(start, end);
                find = true;
            }//if
        }//if
        MetaCatUtil.debugMessage("The docid from online url is:" + docid, 30);
        return docid.trim();
    }


    /**
     * Eocgorid identifier will look like: ecogrid://knb/tao.1.1
     * The AccessionNumber tao.1.1 will be returned. If the given doesn't
     * contains ecogrid, null will be returned.
     * @param identifier String
     * @return String
     */
    public static String getAccessionNumberFromEcogridIdentifier(String identifier)
    {
      String accessionNumber = null;
      if (identifier != null && identifier.startsWith(DBSAXHandler.ECOGRID))
      {
        // find the last "/" in identifier
        int indexOfLastSlash = identifier.lastIndexOf("/");
        int start = indexOfLastSlash+1;
        int end   = identifier.length();
        accessionNumber = identifier.substring(start, end);
      }
      MetaCatUtil.debugMessage("The accession number from url is " +
                                 accessionNumber, 10);
      return accessionNumber;
    }

    private static int getIndexForGivenChar(String str, char character)
    {
        int index = -1;
        // make sure str is not null
        if (str == null) {
            MetaCatUtil.debugMessage(
                    "The given str is null and -1 will be returned", 30);
            return index;
        }
        // got though the string
        for (int i = 0; i < str.length(); i++) {
            // find the first one then break the loop
            if (str.charAt(i) == character) {
                index = i;
                break;
            }//if
        }//for
        MetaCatUtil.debugMessage("the index for char " + character + " is: "
                + index, 30);
        return index;
    }

    /**
     * Utility method to get docid from a given string
     *
     * @param string, the given string should be these two format: 1) str1.str2
     *            in this case docid= str1.str2 2) str1.str2.str3, in this case
     *            docid =str1.str2
     * @param the sperator char
     */
    public static String getDocIdFromString(String str)
    {
        String docId = null;
        if (str == null) {
            MetaCatUtil.debugMessage(
                    "The given str is null and null will be returned"
                            + " in getDocIdfromString", 30);
            return docId;
        } //make sure docid is not null
        int dotNumber = 0;//count how many dots in given string
        int indexOfLastDot = 0;

        //assume that seperator is one charactor string
        char seperator = getOption("accNumSeparator").charAt(0);

        for (int i = 0; i < str.length(); i++) {
            if (str.charAt(i) == seperator) {
                dotNumber++;//count how many dots
                indexOfLastDot = i;//keep the last dot postion
            }
        }//for

        //The string formatt is wrong, because it has more than two or less
        // than
        //one seperator
        if (dotNumber > 2 || dotNumber < 1) {
            docId = null;
        } else if (dotNumber == 2) //the case for str1.str2.str3
        {
            docId = str.substring(0, indexOfLastDot);
        } else if (dotNumber == 1) //the case for str1.str2
        {
            docId = str;
        }

        return docId;
    }//getDocIdFromString

    /**
     * Utility method to get version number from a given string
     *
     * @param string, the given string should be these two format: 1)
     *            str1.str2(no version) version =-1; 2) str1.str2.str3, in this
     *            case version = str3; 3) other, vresion =-2
     */
    public static int getVersionFromString(String str)
            throws NumberFormatException
    {
        int version = -1;
        String versionString = null;
        int dotNumber = 0;//count how many dots in given string
        int indexOfLastDot = 0;

        //assume that seperator is one charactor string
        char seperator = getOption("accNumSeparator").charAt(0);

        for (int i = 0; i < str.length(); i++) {
            if (str.charAt(i) == seperator) {
                dotNumber++;//count how many dots
                indexOfLastDot = i;//keep the last dot postion
            }
        }//for

        //The string formatt is wrong, because it has more than two or less
        // than
        //one seperator
        if (dotNumber > 2 || dotNumber < 1) {
            version = -2;
        } else if (dotNumber == 2 && (indexOfLastDot != (str.length() - 1)))
        //the case for str1.str2.str3
        {
            versionString = str.substring((indexOfLastDot + 1), str.length());
            version = Integer.parseInt(versionString);
        } else if (dotNumber == 1) //the case for str1.str2
        {
            version = -1;
        }

        return version;
    }//getVersionFromString

    /**
     * Utility method to get version string from a given string
     *
     * @param string, the given string should be these two format: 1)
     *            str1.str2(no version) version=null; 2) str1.str2.str3, in
     *            this case version = str3; 3) other, vresion =null;
     */
    public static String getRevisionStringFromString(String str)
            throws NumberFormatException
    {
        // String to store the version
        String versionString = null;
        int dotNumber = 0;//count how many dots in given string
        int indexOfLastDot = 0;

        //assume that seperator is one charactor string
        char seperator = getOption("accNumSeparator").charAt(0);

        for (int i = 0; i < str.length(); i++) {
            if (str.charAt(i) == seperator) {
                dotNumber++;//count how many dots
                indexOfLastDot = i;//keep the last dot postion
            }
        }//for

        //The string formatt is wrong, because it has more than two or less
        // than
        //one seperator
        if (dotNumber > 2 || dotNumber < 1) {
            versionString = null;
        } else if (dotNumber == 2 && (indexOfLastDot != (str.length() - 1))) {
            //the case for str1.str2.str3
            // indexOfLastDot != (str.length() -1) means get rid of str1.str2.
            versionString = str.substring((indexOfLastDot + 1), str.length());
        } else if (dotNumber == 1) //the case for str1.str2 or str1.str2.
        {
            versionString = null;
        }

        return versionString;
    }//getVersionFromString

    /**
     * This method will get docid from an AccessionNumber. There is no
     * assumption the accessnumber will be str1.str2.str3. It can be more. So
     * we think the docid will be get rid of last part
     */
    public static String getDocIdFromAccessionNumber(String accessionNumber)
    {
        String docid = null;
        if (accessionNumber == null) { return docid; }
        String seperator = getOption("accNumSeparator");
        int indexOfLastSeperator = accessionNumber.lastIndexOf(seperator);
        docid = accessionNumber.substring(0, indexOfLastSeperator);
        MetaCatUtil.debugMessage("after parsing accessionnumber, docid is "
                + docid, 30);
        return docid;
    }

    /**
     * This method will get inline data id without the revision number.
     * So if inlineData.1.2 is passed as input, inlineData.2 is returned.
     */
    public static String getInlineDataIdWithoutRev(String accessionNumber)
    {
        String docid = null;
        if (accessionNumber == null) { return docid; }
        String seperator = getOption("accNumSeparator");
        int indexOfLastSeperator = accessionNumber.lastIndexOf(seperator);
        String version = accessionNumber.substring(indexOfLastSeperator,
                                                   accessionNumber.length());
        accessionNumber = accessionNumber.substring(0, indexOfLastSeperator);
        indexOfLastSeperator = accessionNumber.lastIndexOf(seperator);
        docid = accessionNumber.substring(0, indexOfLastSeperator) + version;
        MetaCatUtil.debugMessage("after parsing accessionnumber, docid is "
                                 + docid, 30);

        return docid;
    }

    /**
     * This method will call both getDocIdFromString and
     * getDocIdFromAccessionNumber. So first, if the string looks str1.str2,
     * the docid will be str1.str2. If the string is str1.str2.str3, the docid
     * will be str1.str2. If the string is str1.str2.str3.str4 or more, the
     * docid will be str1.str2.str3. If the string look like str1, null will be
     * returned
     *
     */
    public static String getSmartDocId(String str)
    {
        String docid = null;
        //call geDocIdFromString first.
        docid = getDocIdFromString(str);
        // If docid is null, try to call getDocIdFromAccessionNumber
        // it will handle the seperator more than2
        if (docid == null) {
            docid = getDocIdFromAccessionNumber(str);
        }
        MetaCatUtil.debugMessage("The docid get from smart docid getor is "
                + docid, 30);
        return docid;
    }

    /**
     * This method will get revision from an AccessionNumber. There is no
     * assumption the accessnumber will be str1.str2.str3. It can be more. So
     * we think the docid will be get rid of last part
     */
    public static int getRevisionFromAccessionNumber(String accessionNumber)
            throws NumberFormatException
    {
        String rev = null;
        int revNumber = -1;
        if (accessionNumber == null) { return revNumber; }
        String seperator = getOption("accNumSeparator");
        int indexOfLastSeperator = accessionNumber.lastIndexOf(seperator);
        rev = accessionNumber.substring(indexOfLastSeperator + 1,
                accessionNumber.length());
        revNumber = Integer.parseInt(rev);
        MetaCatUtil.debugMessage("after parsing accessionnumber, rev is "
                + revNumber, 30);
        return revNumber;
    }

    /**
     * Method to get the name of local replication server
     */
    public static String getLocalReplicationServerName()
    {
        String replicationServerName = null;
        String serverHost = null;
        serverHost = getOption("server");
        // append "context/servelet/replication" to the host name
        replicationServerName = serverHost + getOption("replicationpath");
        return replicationServerName;

    }

    /**
     * Method to get docidwithrev from eml2 inline data id The eml inline data
     * id would look like eml.200.2.3
     */
    public static String getDocIdWithoutRevFromInlineDataID(String inlineDataID)
    {
        String docidWithoutRev = null;
        if (inlineDataID == null) { return docidWithoutRev; }
        String seperator = MetaCatUtil.getOption("accNumSeparator");
        char charSeperator = seperator.charAt(0);
        int targetNumberOfSeperator = 2;// we want to know his index
        int numberOfSeperator = 0;
        for (int i = 0; i < inlineDataID.length(); i++) {
            // meet seperator, increase number of seperator
            if (inlineDataID.charAt(i) == charSeperator) {
                numberOfSeperator++;
            }
            // if number of seperator reach the target one, record the index(i)
            // and get substring and terminate the loop
            if (numberOfSeperator == targetNumberOfSeperator) {
                docidWithoutRev = inlineDataID.substring(0, i);
                break;
            }
        }

        MetaCatUtil.debugMessage("Docid without rev from inlinedata id: "
                + docidWithoutRev, 35);
        return docidWithoutRev;

    }

    /**
     * Revise stack change a stack to opposite order
     */
    public static Stack reviseStack(Stack stack)
    {
        Stack result = new Stack();
        // make sure the parameter is correct
        if (stack == null || stack.isEmpty()) {
            result = stack;
            return result;
        }

        while (!stack.isEmpty()) {
            Object obj = stack.pop();
            result.push(obj);
        }
        return result;
    }

    /** A method to replace whitespace in url */
    public static String replaceWhiteSpaceForURL(String urlHasWhiteSpace)
    {
        StringBuffer newUrl = new StringBuffer();
        String whiteSpaceReplace = "%20";
        if (urlHasWhiteSpace == null || urlHasWhiteSpace.trim().equals("")) { return null; }

        for (int i = 0; i < urlHasWhiteSpace.length(); i++) {
            char ch = urlHasWhiteSpace.charAt(i);
            if (!Character.isWhitespace(ch)) {
                newUrl.append(ch);
            } else {
                //it is white sapce, replace it by %20
                newUrl = newUrl.append(whiteSpaceReplace);
            }

        }//for
        MetaCatUtil.debugMessage("The new string without space is:"
                + newUrl.toString(), 35);
        return newUrl.toString();

    }// replaceWhiteSpaceForUR

}
